'''''''''''''''''''''''''''''''''''''''''''''	
' Jos: AD-Doc
' VB-script JoseExec.vbs to perform the documentation
' Version 3.1 beta / 29.06.2011
' Tool for documentation of Active Directory
' Enumerates the main AD objects graphically
' in an HTML file to.
'
' Nils Kaczenski, Ansgar Wiechers, and Nils Weinhold
'
' No guarantee! Use at your own risk!
'
' Recent changes: see changelog.txt
'''''''''''''''''''''''''''''''''''''''''''''

Option Explicit

Const AppName    = "Jos AD-Documentation"
Const AppVersion = "3.1"
Const Authors    = "Nils Kaczenski, Ansgar Wiechers and Nils Weinhold"
Const Homepage   = "http://www.faq-o-matic.net/"

Const ForReading         = 1
Const TristateUseDefault = -2

Const KEY   = 0
Const VALUE = 1

' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
Const ADS_UF_ACCOUNTDISABLE = &h00000002
Const ADS_UF_LOCKOUT        = &h00000010

' See <http://msdn.microsoft.com/en-us/library/cc232505.aspx>
Const GPO_IGNORE  = &h00000001
Const GPO_ENFORCE = &h00000002

Const NTDSDSA_OPT_IS_GC = 1

' See <http://msdn.microsoft.com/en-us/library/cc245476(PROT.13).aspx>
Const DOMAIN_PASSWORD_COMPLEX = &h00000001
Const DOMAIN_PASSWORD_NO_CLEAR_CHANGE = &h00000004
Const DOMAIN_PASSWORD_STORE_CLEARTEXT = &h00000010

Const ResourceDir      = "../Dateien/"
Const ReportStyleSheet = "report.css"
Const ReportDir        = "Reports"
Const DefinitionDir    = "Definitions"
Const ReportSuffix     = ".htm"
Const IndentionStep    = 20 'px ("Tab" width for indention)

' Global variables
Dim fso : Set fso = CreateObject("Scripting.FileSystemObject")

Dim rootDSE
Dim adoCmd
Dim hasADSIErrorMessages
Dim systemBias

Dim showPermissionHint
Dim PermissionReason : PermissionReason = Array()

Dim arrUserAttr
Dim arrComputerAttr
Dim arrGroupAttr
Dim arrContactAttr
Dim arrPrinterAttr

' default configuration values
Dim cfg : Set cfg = CreateObject("Scripting.Dictionary")
	cfg.Add "title", Null
	cfg.Add "reportFilename", Null
	cfg.Add "definitionFilename", Null
	cfg.Add "generator", AppName & " v" & AppVersion
	cfg.Add "includeDomainDetails", False
	cfg.Add "includeTrusts", False
	cfg.Add "includeOUs", False
	cfg.Add "includeUsers", False
	cfg.Add "includeComputers", False
	cfg.Add "includeGroups", False
	cfg.Add "includeContacts", False
	cfg.Add "includePrinters", False
	cfg.Add "includeGPOs", False
	cfg.Add "includeCounts", False
	cfg.Add "includeDates", False
	cfg.Add "includePSOs", False
	cfg.Add "showDescriptions", False
	cfg.Add "showFriendlyNames", False
	cfg.Add "showNaturalNames", False
	cfg.Add "showCaption", False
	cfg.Add "objectFilter", ""
	cfg.Add "debugEnabled", False

' See <http://msdn.microsoft.com/en-us/library/ms680325(VS.85).aspx>
Dim trustFlags : Set trustFlags = CreateObject("Scripting.Dictionary")
	trustFlags.Add &h00000001, "TRUST_ATTRIBUTE_NON_TRANSITIVE" ' Disallow transitivity
	trustFlags.Add &h00000002, "TRUST_ATTRIBUTE_UPLEVEL_ONLY" ' Trust link only valid For uplevel client
	trustFlags.Add &h00000004, "TRUST_ATTRIBUTE_QUARANTINED_DOMAIN" ' Used To quarantine domains
	trustFlags.Add &h00000008, "TRUST_ATTRIBUTE_FOREST_TRANSITIVE" ' This link may contain forest trust information
	trustFlags.Add &h00000010, "TRUST_ATTRIBUTE_CROSS_ORGANIZATION" ' This trust Is To a domain/forest which Is Not part of this enterprise
	trustFlags.Add &h00000020, "TRUST_ATTRIBUTE_WITHIN_FOREST" ' Trust Is internal To this forest
	trustFlags.Add &h00000040, "TRUST_ATTRIBUTE_TREAT_AS_EXTERNAL" ' Trust Is To be treated As external For trust boundary purposes
	trustFlags.Add &h00000080, "TRUST_ATTRIBUTE_TRUST_USES_RC4_ENCRYPTION" ' MIT trust With RC4
	' Trust attributes 0x00000040 through 0x00200000 are reserved For future use
	' Trust attributes 0x00400000 through 0x00800000 were used previously (up To W2K) And should Not be re-used
	' Trust attributes 0x01000000 through 0x80000000 are reserved For user	
	trustFlags.Add &hFF03FFFF, "TRUST_ATTRIBUTES_VALID" ' bit mask to determine if trust attributes are valid
	trustFlags.Add &hFF000000, "TRUST_ATTRIBUTES_USER" ' bit mask to determine if user attributes are set

' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
' corrections inspired by Philipp Fckeler
Dim uacFlags : Set uacFlags = CreateObject("Scripting.Dictionary")
	uacFlags.Add &h00000001, "ADS_UF_SCRIPT"
	uacFlags.Add &h00000002, "ADS_UF_ACCOUNTDISABLE"
	uacFlags.Add &h00000008, "ADS_UF_HOMEDIR_REQUIRED"
	uacFlags.Add &h00000010, "ADS_UF_LOCKOUT"
	uacFlags.Add &h00000020, "ADS_UF_PASSWD_NOTREQD"
	uacFlags.Add &h00000040, "ADS_UF_PASSWD_CANT_CHANGE"
	uacFlags.Add &h00000080, "ADS_UF_ENCRYPTED_TEXT_PASSWORD_ALLOWED"
	uacFlags.Add &h00000100, "ADS_UF_TEMP_DUPLICATE_ACCOUNT"
	uacFlags.Add &h00000200, "ADS_UF_NORMAL_ACCOUNT"
	uacFlags.Add &h00000800, "ADS_UF_INTERDOMAIN_TRUST_ACCOUNT"
	uacFlags.Add &h00001000, "ADS_UF_WORKSTATION_TRUST_ACCOUNT"
	uacFlags.Add &h00002000, "ADS_UF_SERVER_TRUST_ACCOUNT"
	uacFlags.Add &h00010000, "ADS_UF_DONT_EXPIRE_PASSWD"
	uacFlags.Add &h00020000, "ADS_UF_MNS_LOGON_ACCOUNT"
	uacFlags.Add &h00040000, "ADS_UF_SMARTCARD_REQUIRED"
	uacFlags.Add &h00080000, "ADS_UF_TRUSTED_FOR_DELEGATION"
	uacFlags.Add &h00100000, "ADS_UF_NOT_DELEGATED"
	uacFlags.Add &h00200000, "ADS_UF_USE_DES_KEY_ONLY"
	uacFlags.Add &h00400000, "ADS_UF_DONT_REQUIRE_PREAUTH"
	uacFlags.Add &h00800000, "ADS_UF_PASSWORD_EXPIRED"
	uacFlags.Add &h01000000, "ADS_UF_TRUSTED_TO_AUTHENTICATE_FOR_DELEGATION"
	uacFlags.Add &h02000000, "ADS_ UF_NO_AUTH_DATA_REQUIRED"
	uacFlags.Add &h04000000, "ADS_ UF_PARTIAL_SECRETS_ACCOUNT"

' Dictionary for translating attribute names into (user-)friendly names.
Dim naturalName : Set naturalName = CreateObject("Scripting.Dictionary")
	naturalName.Add "altrecipient", "Mail Forwarding"
	naturalName.Add "canonicalname", "OU-Path (canonicalName)"
	naturalName.Add "company", "Company"
	naturalName.Add "dcroles", "Computer Role"
	naturalName.Add "department", "Department"
	naturalName.Add "distinguishedname", "LDAP Name (DN)"
	naturalName.Add "e-mail adressenickname", "Mail Alias"
	naturalName.Add "givenname", "First Name"
	naturalName.Add "grouptype", "Group type"
	naturalName.Add "homedirectory", "Local Path"
	naturalName.Add "homedrive", "Home Drive"
	naturalName.Add "location", "Location"
	naturalName.Add "logonhours", "Logon hours"
	naturalName.Add "mail", "E-Mail Address"
	naturalName.Add "member", "Member"
	naturalName.Add "memberof", "Member Of"
	naturalName.Add "msds-psoapplied", "Valid PSOs"
	naturalName.Add "msds-resultantpso", "Effective PSO"
	naturalName.Add "msexchhomeservername", "Exchange Homeserver"
	naturalName.Add "mstshomedirectory", "Terminalserver Home (msTS*)"   ' Server 2008 only - unused! (see: <http://msdn.microsoft.com/en-us/library/cc220562.aspx>)
	naturalName.Add "mstsprofilepath", "Terminalserver Profil (msTS*)"   ' Server 2008 only - unused! (see: <http://msdn.microsoft.com/en-us/library/cc220579.aspx>)
	naturalName.Add "objectsid", "SID"
	naturalName.Add "operatingsystem", "Operating System"
	naturalName.Add "operatingsystemservicepack", "Service Pack"
	naturalName.Add "operatingsystemversion", "Operating system version"
	naturalName.Add "physicaldeliveryofficename", "Office"
	naturalName.Add "primarygroupid", "Primary Group (ID)"
	naturalName.Add "primarygrouptoken", "Token of the Primary Group"
	naturalName.Add "printsharename", "Share Name"
	naturalName.Add "profilepath", "Roaming Profile"
	naturalName.Add "proxyaddresses", "Exchange E-Mail Addresses"
	naturalName.Add "pwdlastset", "Password changed"
	naturalName.Add "samaccountname", "Logon Name"
	naturalName.Add "scriptpath", "Logon Script"
	naturalName.Add "servername", "Print Server"
	naturalName.Add "showinaddressbook", "Display in Address Book"
	naturalName.Add "sidhistory", "SID history"
	naturalName.Add "sn", "Last Name"
	naturalName.Add "streetaddress", "Address"
	naturalName.Add "telephonenumber", "Telephone"
	naturalName.Add "terminalserviceshomedirectory", "Terminalserver Home (userParameters)" 
	naturalName.Add "terminalservicesprofilepath", "Terminalserver Profil (userParameters)" 
	naturalName.Add "useraccountcontrol", "Accounts attribute"
	naturalName.Add "userprincipalname", "UPN Logon Name"
	naturalName.Add "whenchanged", "Object modified on"
	naturalName.Add "whencreated", "Object created on"

' Symbol tags
Dim symbols : Set symbols = CreateObject("Scripting.Dictionary")
	symbols.Add "Domain", "<img src=""" & ResourceDir & "domain.gif"">&nbsp;"
	symbols.Add "organizationalUnit", "<img src=""" & ResourceDir & "ou.gif"">&nbsp;"
	symbols.Add "container", "<img src=""" & ResourceDir & "ordner.gif"">&nbsp;"
	symbols.Add "builtinDomain", "<img src=""" & ResourceDir & "ordner.gif"">&nbsp;"
	symbols.Add "User", "<img src=""" & ResourceDir & "user.gif"">&nbsp;"
	symbols.Add "Computer", "<img src=""" & ResourceDir & "computer.gif"">&nbsp;"
	symbols.Add "Group", "<img src=""" & ResourceDir & "group.gif"">&nbsp;"
	symbols.Add "Contact", "<img src=""" & ResourceDir & "contact.gif"">&nbsp;"
	symbols.Add "Printer", "<img src=""" & ResourceDir & "printer.gif"">&nbsp;"
	symbols.Add "GPO", "<img src=""" & ResourceDir & "gpo.gif"">&nbsp;"
	symbols.Add "PSO", "<img src=""" & ResourceDir & "pso.gif"">&nbsp;"
	symbols.Add "disabled", "<img src=""" & ResourceDir & "disable.gif"">&nbsp;"
	symbols.Add "locked", "<img src=""" & ResourceDir & "locked.gif"">&nbsp;"
	symbols.Add "Question", "<img src=""" & ResourceDir & "question.gif"">&nbsp;"
	symbols.Add "Exclamation", "<img src=""" & ResourceDir & "exclamation.gif"">&nbsp;"
	symbols.Add "blocked", "<img src=""" & ResourceDir & "blocked.gif"">&nbsp;"
	symbols.Add "Clock", "<img src=""" & ResourceDir & "uhr.gif"">&nbsp;"
	symbols.Add "Jos", "<img src=""" & ResourceDir & "jose-klein.gif"">"
	symbols.Add "FOM", "<img src=""" & ResourceDir & "fom-klein.gif"">"
Dim stats : Set stats = CreateObject("Scripting.Dictionary")
	stats.Add "user", 0
	stats.Add "computer", 0
	stats.Add "printQueue", 0
	stats.Add "group", 0
	stats.Add "contact", 0
	stats.Add "organizationalUnit", 0
	' we don't use container in v3.1 any more
	' stats.Add "container", 0


Main WScript.Arguments


Sub Main(args)
	Dim domain
	
	If InStr(LCase(WScript.FullName), "cscript") = 0 Then
		MsgBox "This script can be started not with WScript.exe." & vbNewLine _
			& "Please launch it from the command line with the following command:" & vbNewLine _
			& vbNewLine & "cscript """ & WScript.ScriptFullName & """ /?" _
			, vbInformation, WScript.ScriptName & " - Start with CScript.exe"
		WScript.Quit 1
	End If
	
	If WScript.Version < "5.6" Then
		WScript.Echo "This script requires at least the Windows script host 5.6." & vbNewLine _
			& "Is the version of your WSH: " & WScript.Version & vbNewLine _
			& "Please install the latest version, which you can" & vbNewLine _
			& "download at Microsoft."
		WScript.Quit 1
	End If

	If args.count = 0 Then
		PrintSyntax
	End If

	On Error Resume Next
	Set rootDSE = GetObject("LDAP://rootDSE")
	domain = rootDSE.Get("DefaultNamingContext")
	If TrapError("No domain available") Then WScript.Quit 1
	On Error Goto 0

	systemBias = GetSystemBias
	hasADSIErrorMessages = CheckForADSIErrorMessages
	showPermissionHint = False

	' Open AD connection only once.
	Dim adoConn : Set adoConn = CreateObject("ADODB.Connection")
	adoConn.Provider = "ADsDSOObject"
	adoConn.Open "Active Directory Provider"

	Set adoCmd = CreateObject("ADODB.Command")
	adoCmd.ActiveConnection = adoConn
	adoCmd.Properties("Page Size") = 100
	adoCmd.Properties("Timeout") = 30
	adoCmd.Properties("Cache Results") = False

	ParseCmd(args)
	ReadDefinitions cfg("definitionFilename")

	CreateReport domain

	adoConn.Close
	Set adoConn = Nothing
	Set adoCmd = Nothing
End Sub

' ==============================================================================
' Configuration functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Parse the commandline arguments.
' ------------------------------------------------------------------------------
Sub ParseCmd(args)
	If args.Named.Exists("?") Then
		PrintSyntax
	Else
		With args.Named
			' Definition filename (optional)
			' If filename is "pipe" or is omitted, definitions will be read
			' from STDIN.
			If .Exists("d") And .Item("d") <> "" Then
				If LCase(.Item("d")) <> "pipe" Then cfg("definitionFilename") = .Item("d")
			End If
			' note: should there be no title or report file name they will be generated in CreateReport!
			' Report filename (optional)
			If .Exists("r") And .Item("r") <> "" Then 
				cfg("reportFilename") = .Item("r")
			End If
			' Report title (optional)
			If .Exists("t") And .Item("t") <> "" Then cfg("title") = .Item("t")
			' Object filter (optional)
			If .Exists("f") And .Item("f") <> "" Then cfg("objectFilter") = .Item("f")
			If .Exists("debug") Then cfg("debugEnabled") = True
		End With
	End If

	PrintDebug "Definitionsdatei: " & cfg("definitionFilename")
	PrintDebug "Reportdatei: " & cfg("reportFilename")
	PrintDebug "Titel: " & cfg("title")
	PrintDebug "Filter: " & cfg("objectFilter")
	PrintDebug "System Time Zone Bias: " & GetSystemBias

End Sub

' ------------------------------------------------------------------------------
' Read definitions of what to include in the report.
' ------------------------------------------------------------------------------
Sub ReadDefinitions(filename)
	Dim defFile, objFile, line, section
	Dim userAttribs, computerAttribs, groupAttribs, contactAttribs, printerAttribs

	Dim userAttributes     : Set userAttributes = CreateAttributeSet()
	Dim computerAttributes : Set computerAttributes = CreateAttributeSet()
	Dim groupAttributes    : Set groupAttributes = CreateAttributeSet()
	Dim contactAttributes  : Set contactAttributes = CreateAttributeSet()
	Dim printerAttributes  : Set printerAttributes = CreateAttributeSet()

	' Read definitions from STDIN unless a definition file is specified.
	If IsNull(filename) Then
		Set objFile = WScript.StdIn
		PrintDebug "Read definitions from STDIN ..."
	Else
		defFile = fso.BuildPath(fso.BuildPath(fso.GetParentFolderName(WScript.ScriptFullName), DefinitionDir), filename)
		If Not fso.FileExists(defFile) Then
			WScript.StdErr.WriteLine "Definition file '" & defFile & "' does not exist!"
			WScript.Quit 1
		End If
		Set objFile = fso.OpenTextFile(defFile, ForReading)
		PrintDebug "Read definitions from '" & defFile & "' ..."
	End If

	section = ""
	Do Until objFile.AtEndOfStream
		line = Trim(Replace(objFile.ReadLine, vbTab, " "))

		If Left(line, 1) = "[" Then
			PrintDebug line
			section = line
			Select Case LCase(section)
				Case "[user]" cfg("includeUsers") = True
				Case "[computer]" cfg("includeComputers") = True
				Case "[group]" cfg("includeGroups") = True
				Case "[contact]" cfg("includeContacts") = True
				Case "[printer]" cfg("includePrinters") = True
			End Select
		ElseIf line = "" Or Left(line, 1) = ";" Then
			' continue (line is either blank or a comment)
		Else
			PrintDebug line
			Select Case LCase(section)
				Case "[meta]"
					Dim def : def = EvalDefinition(line)
					' Use report title and filename from definition file only if they
					' haven't been set already via commandline.
					Select Case LCase(def(KEY))
						' note: should there be no title or report file name they will be generated below!
						Case "reportname" If IsNull(cfg("title")) Then cfg("title") = def(VALUE)
						Case "reportfilename" If IsNull(cfg("reportFilename")) Then cfg("reportFilename") = def(VALUE)
						Case "objectfilter" cfg("objectFilter") = def(VALUE)
						Case "showsymbols" cfg("showCaption") = def(VALUE)
						Case "shownaturalname" cfg("showNaturalNames") = def(VALUE)
						Case "showfriendlyname" cfg("showFriendlyNames") = def(VALUE)
						' Case "showdescriptions" cfg("showDescriptions") = def(VALUE)
						Case Else WScript.StdErr.WriteLine "Unknown definition in " & section & ": " & line
					End Select
				Case "[common]"
					Select Case LCase(line)
						Case "fsmo" cfg("includeDomainDetails") = True
						Case "trust" cfg("includeTrusts") = True
						Case "objectreport" cfg("includeOUs") = True
						Case "gpo" cfg("includeGPOs") = True
						Case "number" cfg("includeCounts") = True
						Case "modify" cfg("includeDates") = True
						Case "description" cfg("showDescriptions") = True
						Case "pso" cfg("includePSOs") = True
						Case "permission" cfg("showAdvancedPermissionHint") = True
						Case Else WScript.StdErr.WriteLine "Unknown definition in " & section & ": " & line
					End Select
				Case "[user]"
					If Not userAttributes.Exists(line) Then userAttributes.Add line, True
					' If properties for MSTS are included, add versions both for
					' Windows Server 2003 and 2008 as they use different attributes
					If LCase(line) = "terminalserviceshomedirectory" Then
						userAttributes.Add "mstshomedirectory", True
					End If
					If LCase(line) = "terminalservicesprofilepath" Then
						userAttributes.Add "mstsprofilepath", True
					End If
				Case "[computer]"
					If Not computerAttributes.Exists(line) Then computerAttributes.Add line, True
				Case "[group]"
					If Not groupAttributes.Exists(line) Then groupAttributes.Add line, True
				Case "[contact]"
					If Not contactAttributes.Exists(line) Then contactAttributes.Add line, True
				Case "[printer]"
					If Not printerAttributes.Exists(line) Then printerAttributes.Add line, True
				Case Else WScript.StdErr.WriteLine "Unknown section name: " & section
			End Select
		End If
	Loop

	if IsNull(filename) Then objFile.Close ' don't close STDIN
	Set objFile = Nothing
	
	' add Password Setting Object attributes for users if needed
	If cfg("includePSOs") Then 
		userAttributes.Add "msDS-PSOApplied", True
		userAttributes.Add "msDS-ResultantPSO", True
	End If

	' Set report title and filename to default values if they haven't been
	' set via commandline or definition file.
	If IsEmpty(cfg("title")) Then cfg("title") = DefaultTitle & " " & CStr(Now)
	If IsEmpty(cfg("reportFilename")) Then cfg("reportFilename") = Replace(Replace(cfg("title"), ":", "-"), "/", "-") & ReportSuffix

	arrUserAttr = userAttributes.Keys
	arrComputerAttr = computerAttributes.Keys
	arrGroupAttr = groupAttributes.Keys
	arrContactAttr = contactAttributes.Keys
	arrPrinterAttr = printerAttributes.Keys

	Set userAttributes = Nothing
	Set computerAttributes = Nothing
	Set groupAttributes = Nothing
	Set contactAttributes = Nothing
	Set printerAttributes = Nothing

	PrintDebug "Current Definitions:"
	Dim setting
	For Each setting In cfg.Keys
		PrintDebug setting & ":" & vbTab & cfg(setting)
	Next

	PrintDebug "User attributes: " & Join(arrUserAttr, ",")
	PrintDebug "Computer attributes: " & Join(arrComputerAttr, ",")
	PrintDebug "Group attributes: " & Join(arrGroupAttr, ",")
	PrintDebug "Contact attributes: " & Join(arrContactAttr, ",")
	PrintDebug "Printer attributes: " & Join(arrPrinterAttr, ",")
End Sub

' ==============================================================================
' Report generation functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Create the report for the given domain.
' ------------------------------------------------------------------------------
Sub CreateReport(domain)
	Dim reportFile, objReport, objDomain, folder, policies, domainInfo

	PrintDebug "Create report " & cfg("reportFilename") & " ..."

	folder = fso.BuildPath(fso.GetParentFolderName(WScript.ScriptFullName), ReportDir)
	If Not fso.FolderExists(folder) Then
		fso.CreateFolder folder
		PrintDebug "Created folder: " & folder
	End If

	On Error Resume Next
	reportFile = fso.BuildPath(folder, cfg("reportFilename"))
	Set objReport = fso.CreateTextFile(reportFile)
	If TrapError("Error creating file '" & reportFile & "'") Then WScript.Quit 1
	On Error Goto 0

	' write HTML header
	objReport.WriteLine "<!DOCTYPE HTML PUBLIC ""-//W3C//DTD HTML 4.01 Transitional//EN" & vbNewLine _
		& vbTab & """http://www.w3.org/TR/html4/loose.dtd"">" & vbNewLine _
		& "<HTML>" & vbNewLine & "<HEAD>" & vbNewLine _
		& "<TITLE>" & cfg("title") & "</TITLE>" & vbNewLine _
		& "<META NAME=""generator"" CONTENT=""" & cfg("generator") _
		& " (von " & Authors & ")"">" & vbNewLine _
		& "<LINK REL=""stylesheet"" TYPE=""text/css"" HREF=""" & ResourceDir _
		& ReportStyleSheet & """>" & vbNewLine & "</HEAD>" & vbNewLine & "<BODY>" _
		& vbNewLine & "<H1 id=""title"">" & cfg("title") & "</H1>"
	If cfg("objectFilter") <> "" Then
		objReport.WriteLine "<P class=""filter"">Edition limited to elements of the Container: <em>" & cfg("objectFilter") & "</em></P>"
	End If

	If cfg("includeGPOs") Then 
		Set policies = GetGPOs(domain)
	Else
		Set policies = Nothing
	End If

	' we need to read the domain data in every case as we may need the details later
	Set domainInfo = GetDomainInfo(domain)

	WriteDomainInfo objReport, domainInfo
	If cfg("includeTrusts") Then WriteTrusts objReport, GetTrusts(domain)
	if cfg("includePSOs") then WritePSOs objReport, GetPSOs(domain)
	If cfg("includeOUs") Then
		WriteOUInfo objReport, GetOUs("builtinDomain", domain, 0), policies, domainInfo("lockoutDuration")
		WriteOUInfo objReport, GetOUs("container", domain, 0), policies, domainInfo("lockoutDuration")
		WriteOUInfo objReport, GetOUs("organizationalUnit", domain, 0), policies, domainInfo("lockoutDuration")
	End If
	If cfg("includeGPOs") Then WriteGPOInfo objReport, policies
	If cfg("showCaption") Then WriteCaption objReport
	WriteAdv objReport

	Set policies = Nothing
	Set domainInfo = Nothing

	' write HTML footer
	objReport.WriteLine "</BODY>" & vbNewLine & "</HTML>"

	PrintDebug "Finished."

	objReport.Close
	Set objReport = Nothing
End Sub

' ------------------------------------------------------------------------------
' Write information of the given domain object to the report.
' ------------------------------------------------------------------------------
Sub WriteDomainInfo(report, domainInfo)
	Dim i
	
	report.Write "<strong>" & symbols("Domain") & domainInfo("fqdn")
	WriteDescription report, domainInfo("description")
	report.WriteLine "</strong>"

	If Not cfg("includeDomainDetails") Then Exit Sub

	report.WriteLine "<TABLE style=""margin-left:" & IndentionStep & "px;""><TR>"

	If cfg("includeDates") Then
		report.WriteLine "<TR><TD>Object created on</TD><TD class=""value"">" & domainInfo("whenCreated") & "</TD></TR>"
		report.WriteLine "<TR><TD>Object modified on</TD><TD class=""value"">" & domainInfo("whenChanged") & "</TD></TR>"
	End If

	report.WriteLine "<TR><TD>Last backed up</TD><TD class=""value"">" & domainInfo("lastBackup")& "</TD></TR>"
	report.WriteLine "<TR><TD>Queried DC</TD><TD class=""value"">" & domainInfo("DCtoQuery") & "</TD></TR>"
	report.WriteLine "<TR><TD>Root domain (Forest)</TD><TD class=""value"">" & domainInfo("rootDomain") & "</TD></TR>"

	' Write FSMO role owners.
	report.WriteLine "<TR><TD>Schema-Master (Forest)</TD><TD class=""value"">" & domainInfo("fsmoSchemaMaster") & "</TD></TR>"
	report.WriteLine "<TR><TD>DNS-Master (Forest)</TD><TD class=""value"">" & domainInfo("fsmoDnsMaster") & "</TD></TR>"
	report.WriteLine "<TR><TD>PDC-Emulator (Domain)</TD><TD class=""value"">" & domainInfo("fsmoPdcEmulator") & "</TD></TR>"
	report.WriteLine "<TR><TD>RID-Master (Domain)</TD><TD class=""value"">" & domainInfo("fsmoRidMaster") & "</TD></TR>"
	report.WriteLine "<TR><TD>Infrastructure-Master (Domain)</TD><TD class=""value"">" & domainInfo("fsmoInfraMaster") & "</TD></TR>"

	' Write operation modes.
	report.WriteLine "<TR><TD>Forest Functional level</TD><TD class=""value"">" & domainInfo("opModeForest") & "</TD></TR>"
	report.WriteLine "<TR><TD>Domain Functional level</TD><TD class=""value"">" & domainInfo("opModeDomain") & "</TD></TR>"

	' Write schema versions
	report.WriteLine "<TR><TD>Schema Version</TD><TD class=""value"">" & domainInfo("schemaVersion") & "</TD></TR>"
	If domainInfo("exchSchemaVersion") <> "" Then report.WriteLine "<TR><TD>Exchange Schema Version</TD><TD class=""value"">" & domainInfo("exchSchemaVersion") & "</TD></TR>"

	report.WriteLine "<TR><TD>NetBIOS Domainname</TD><TD class=""value"">" & domainInfo("netBiosName") & "</TD></TR>"
	report.WriteLine "<TR><TD>Domain SID</TD><TD class=""value"">" & domainInfo("sID") & "</TD></TR>"

	' Write deleted object settings
	report.WriteLine "<TR><TD>Tombstone Lifetime</TD><TD class=""value"">" 
	If domainInfo("tombstoneLifetime") = 0  Then
		report.Write "60 Tage"
	Else	
		report.Write domainInfo("tombstoneLifetime") & " Day"
	End If
	report.Write "</TD></TR>"
	report.WriteLine "<TR><TD>Deleted Objects Lifetime</TD><TD class=""value"">" 
	If domainInfo("msDS-DeletedObjectLifetime") = 0  Then
		report.Write "(not defined)"
	Else	
		report.Write domainInfo("msDS-DeletedObjectLifetime") & " Days"
	End If
	report.Write "</TD></TR>"
	
	' Write AD Recycle Bin
	report.WriteLine "<TR><TD>AD Recycle Bin</TD><TD class=""value"">"
	If domainInfo("HasRecycleBin") Then
		report.Write "enabled"
	Else
		report.Write "not enabled"
	End If
	report.Write "</TD></TR>"
	
	' Write all DCs
	report.WriteLine "<TR><TD>DCs of the Domain<span class=""value"">(" & UBound(domainInfo("allDCs")) + 1 & ")</span></TD><TD class=""value"">"
	For i = 0 To UBound(domainInfo("allDCs"))
		report.WriteLine domainInfo("allDCs")(i) & "<br />"
	Next
	report.WriteLine "</TD></TR>"
	
	' Write account lockout settings.
	report.WriteLine "<TR><TD>Accounts lock duration</TD><TD class=""value"">" 
	Select Case domainInfo("lockoutDuration") 
		Case -1 report.Write "(Value not readable)"
		Case 0 report.Write "only manual unlock"
		Case Else report.Write domainInfo("lockoutDuration") & " Minutes"
	End Select
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Account lockout after</TD><TD class=""value"">" & domainInfo("lockoutThreshold") & " wrong Passwords"
	If domainInfo("lockoutThreshold") = 0 Then report.Write " (no locks)"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Deleted lockout counter after</TD><TD class=""value"">" & domainInfo("lockoutObservationWindow") & " Minutes"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Minimum Password length</TD><TD class=""value"">" & domainInfo("minPwdLength") & " Characters"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Maximum Password age</TD><TD class=""value"">" & domainInfo("maxPwdAge") & " Day"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Minimum Password age</TD><TD class=""value"">" & domainInfo("minPwdAge") & " Day"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Password History</TD><TD class=""value"">" & domainInfo("pwdHistoryLength") & " old Passwords"
	report.Write "</TD></TR>" & vbNewLine & "<TR><TD>Password properties</TD><TD class=""value"">" & DecodePwdProperties(domainInfo("pwdProperties"))
	report.WriteLine "</TD></TR>" & vbNewLine & "</TABLE>"
	report.WriteLine "<div>&nbsp;</div>"
End Sub

' ------------------------------------------------------------------------------
' Write given trusts to report.
' ------------------------------------------------------------------------------
Sub WriteTrusts(report, arrTrusts)
	Dim i
	For i = LBound(arrTrusts) To UBound(arrTrusts)
		report.WriteLine "<div style=""margin-left:" & IndentionStep & "px;"">" _
			& symbols("Domain") & " Trust Relationship: <I>" & arrTrusts(i)("name") _
			& " (" & arrTrusts(i)("flatName") & ")" & vbNewLine _
			& "<UL style=""margin-left:" & IndentionStep & "px;""><LI>" & arrTrusts(i)("type") & "</LI>"
		If Not IsNull(arrTrusts(i)("direction")) Then report.WriteLine "<LI>" & arrTrusts(i)("direction") & "</LI>"
		If Not IsNull(arrTrusts(i)("attributes")) Then report.WriteLine "<LI>Attribute: " & arrTrusts(i)("attributes") & "</LI>"
		report.WriteLine "</UL></I></div>"
	Next
End Sub

' ------------------------------------------------------------------------------
' Write the given OU data to the report.
' ------------------------------------------------------------------------------
Sub WriteOUInfo(report, ou, allGPOs, lockoutDuration)
	Dim indent1, indent2, category, obj, child

	indent1 = ou("level") * IndentionStep
	indent2 = indent1 + IndentionStep

	PrintDebug "Write " & ou("name") & " data ..."

	' Don't write top-level element, because WriteDomainInfo() already did that.
	If ou("level") > 0 Then
		report.Write "<P style=""margin-left:" & indent1 & "px;"">" _
			& symbols(ou("class")) & "<span class=""OULevel"" title=""OU level " & ou("level") & """>" & ou("level") & "</span>&nbsp;<B>" & ou("name") & "</B>"
		WriteDescription report, ou("description")
		report.WriteLine "</P>"
		If cfg("includeDates") Then
			report.WriteLine "<P style=""margin-left:" & indent2 & "px;"">Object created on: <I>" & ou("whenCreated") & "</I></P>"
			report.WriteLine "<P style=""margin-left:" & indent2 & "px;"">Object modified on: <I>" & ou("whenChanged") & "</I></P>"
		End If
	End If

	If cfg("includeCounts") And ou.Exists("count") And ou("level") > 0 Then WriteCount report, ou("count"), indent2
	If cfg("includeGPOs") And ou.Exists("GPO") Then WriteLinkedGPOInfo report, ou("GPO"), ou("gpoInherit"), allGPOs, indent2
	If cfg("includeUsers") And ou.Exists("User") Then WriteObjectInfo report, "User", ou, lockoutDuration, indent2
	If cfg("includeComputers") And ou.Exists("Computer") Then WriteObjectInfo report, "Computer", ou, Null, indent2
	If cfg("includeGroups") And ou.Exists("Group") Then WriteObjectInfo report, "Group", ou, Null, indent2
	If cfg("includeContacts") And ou.Exists("Contact") Then WriteObjectInfo report, "Contact", ou, Null, indent2
	If cfg("includePrinters") And ou.Exists("Printer") Then WriteObjectInfo report, "Printer", ou, Null, indent2

	For Each child In ou("children")
		WriteOUInfo report, child, allGPOs, lockoutDuration
	Next
End Sub

' ------------------------------------------------------------------------------
' Write object counts to the report.
' ------------------------------------------------------------------------------
Sub WriteCount(report, count, indent)
	Dim category

	If IsEmpty(count) Then Exit Sub
	If count Is Nothing Then Exit Sub
	' should indent be 20 it's the domain level where we don't want object counts
	'If intent = 20 Then Exit Sub

	PrintDebug vbTab & "Write counts ..."

	report.WriteLine "<P style=""margin-left:" & indent _
		& "px;"">Number of objects: " & count("total") & "<UL>"
	If count("total") > 0 Then
		For Each category In count.Keys
			If category <> "total" And count(category) > 0 Then
				report.WriteLine "<LI style=""margin-left:" & indent + IndentionStep _
					& "px;"">Number of objects of type """ & category & """: " _
					& count(category) & "</LI>"
			End If
		Next
	End If
	report.WriteLine "</UL></P>"
End Sub

' ------------------------------------------------------------------------------
' Write GPOs that are linked to the current container to the report.
' ------------------------------------------------------------------------------
Sub WriteLinkedGPOInfo(report, linkedGPOs, inherit, allGPOs, indent)
	Dim gpo

	If UBound(linkedGPOs) < 0 Then Exit Sub
	PrintDebug "Write linked group policies ..."

	If Not inherit Then report.WriteLine "<P style=""margin-left:" & indent _
		& "px;"">" & symbols("blocked") _
		& "Vererbung von Gruppenrichtlinien ist deaktiviert</P>"

	For Each gpo In linkedGPOs
		PrintDebug vbTab & "Linked policy: " & gpo("guid")
		If allGPOs.Exists(gpo("guid")) Then
			' report only if GPO has been collected before
			report.Write "<P style=""margin-left:" & indent & "px;"">" & symbols("GPO")
			If gpo("disabled") Then report.Write symbols("disabled")
			If gpo("enforced") Then report.Write symbols("Exclamation")
			report.WriteLine "<B>" & allGPOs(gpo("guid"))("displayName") _
				& "</B>&nbsp;<I>(" & gpo("guid") & ")</I></P>"
		Else
			' GPO has not been collected - most likely a permission issue
			report.Write "<P style=""margin-left:" & indent & "px;"">" & symbols("GPO") _
				& symbols("Question") & "<B><I>(Access not possible)</B>&nbsp;(" & gpo("guid") & ")</I></P>"
			showPermissionHint = True
			SetPermissionReason "GPO access " & gpo("guid")
		End If

	Next
End Sub

' ------------------------------------------------------------------------------
' Write the data of the given parent object collection to the report.
' ------------------------------------------------------------------------------
Sub WriteObjectInfo(report, category, parentObj, lockoutDuration, indent)
	Dim obj, unlockTime

	If parentObj.Exists("class") Then
		PrintDebug "Write " & LCase(category) & "s of " & parentObj("class") & " " & parentObj("name") & " ..."
	Else
		PrintDebug "Write " & LCase(category) & "s of host " & parentObj("name") & " ..."
	End If

	For Each obj In parentObj(category)
		PrintDebug vbTab & category & ": " & obj("name")
		report.Write "<P style=""margin-left:" & indent & "px;"">" & symbols(category)
		If category = "User" Or category = "Computer" Then
			If IsNull(obj("disabled")) Then
				report.Write symbols("Question")
			ElseIf obj("disabled") Then
				report.Write symbols("disabled")
			End If
		End If
		report.Write "<B>" & obj("name") & "</B>"
		If category = "User" Then
			If Not IsNull(obj("locked")) And obj("locked") And Not IsEmpty(lockoutDuration) Then
				report.Write "&nbsp;" & symbols("locked") & "locked since: " & obj("lockoutTime")
				If IsNull(obj("lockoutTime")) Then
					report.Write "(Value not readable)"
				' lockoutDuration contains -1 when the value is not retrievable
				' in that case we should not calculate the unlock time
				ElseIf lockoutDuration > -1 Then	
					unlockTime = DateAdd("n", lockoutDuration, obj("lockoutTime"))
					If (Now > unlockTime) Then
						report.Write ", unlocked since: "
					Else
						report.Write ", is unlocked: "
					End If
					report.Write unlockTime
				Else
					report.Write " (Unlock not readable)"
				End If
			End If
			If obj("expires") <> #1/1/1601# Then report.Write "&nbsp;" & symbols("Clock") & "Account expires: " & obj("expires") & " "
		End If
		WriteDescription report, obj("description")
		report.WriteLine "</P>"
		WriteAttributes report, obj("attributes"), (indent + IndentionStep)
		If obj.Exists("Printer") And cfg("includePrinters") Then WriteObjectInfo report, "Printer", obj, Null, (indent + IndentionStep)
	Next
End Sub

' ------------------------------------------------------------------------------
' Write the given descriptive comment to the report.
' ------------------------------------------------------------------------------
Sub WriteDescription(report, descr)
	If cfg("showDescriptions") And descr <> "" Then report.Write "&nbsp;<I>(" & descr & ")</I>"
End Sub

' ------------------------------------------------------------------------------
' Write the given attributes to the report.
' ------------------------------------------------------------------------------
Sub WriteAttributes(report, attributes, indent)
	Dim name, value, isNotDN, NumberOfValues

	' do not write a table when no attributes to report on
	If attributes.Count = 2 and not cfg("includeDates") Then Exit Sub

	report.WriteLine "<TABLE style=""margin-left:" & indent & "px;"">"
	For Each name In attributes.Keys
		' Write dates only if includeDates flag is set.
		If Not (LCase(name) = "whencreated" Or LCase(name) = "whenchanged") Or cfg("includeDates") Then
			isNotDN = (name <> "distinguishedName")
			VALUE = attributes(name)
			If cfg("showNaturalNames") Then name = Translate(name)
			If IsArray(value) Then
				Dim i
				' Get number of values (e.g. group members)
				NumberOfValues = UBound(value) + 1
				If NumberOfValues > 1 Then name = name & " (" & NumberOfValues & ")"
				' Walk through values
				For i = LBound(value) To UBound(value)
					WriteSingleAttribute report, name, VALUE(i), isNotDN
					' Don't write attribute name in subsequent lines.
					If i = LBound(value) Then name = "&nbsp;"
				Next
			Else
				WriteSingleAttribute report, name, value, isNotDN
			End If
		End If
	Next
	report.WriteLine "</TABLE>"
End Sub

' ------------------------------------------------------------------------------
' Write a given attribute to the report.
' ------------------------------------------------------------------------------
Sub WriteSingleAttribute(report, ByVal name, ByVal value, isNotDN)
	PrintDebug vbTab & vbTab & name & ": " & VALUE
	If VALUE <> "" Then
		If cfg("showFriendlyNames") And isNotDN And (InStr(1, value, "cn=", vbTextCompare) = 1) Then
			VALUE = Mid(value, 4, InStr(4, value, "=", vbTextCompare) - 7)
		End If
		report.WriteLine "<TR><TD width=""175"">" & name _
			& "</TD><TD class=""value"">" & VALUE & "</TD></TR>"
	End If
End Sub

' ------------------------------------------------------------------------------
' Write the given GPO data to the report.
' ------------------------------------------------------------------------------
Sub WriteGPOInfo(report, gpoList)
	Dim guid, link

	If gpoList.Count = 0 Then Exit Sub

	PrintDebug "Write group policy data ..."

	report.WriteLine "<H1>List of Group Policy (GPOs)</H1>"

	For Each guid In gpoList.Keys
		PrintDebug vbTab & "Policy: " & guid
		report.WriteLine "<P class=""policy"">" & symbols("GPO") & "<strong>" _
			& gpoList(guid)("displayName") & "</strong><BR>" & vbNewLine _
			& "<em>{" & guid & "}</em><BR>"

		If cfg("showDescriptions") And gpoList(guid)("comment") <> "" Then report.WriteLine "Comment:<BR><em>" & vbNewLine & Replace(gpoList(guid)("comment"), vbNewLine, "<BR>") & "</em><BR>"
		If gpoList(guid)("gpcWMIFilter") <> "" Then report.WriteLine "WMI-Filter: " & gpoList(guid)("gpcWMIFilter") & "<BR>"

		If cfg("includeDates") Then
			report.WriteLine "created at: " & gpoList(guid)("whenCreated") & "<BR>"
			report.WriteLine "modified at: " & gpoList(guid)("whenChanged") & "<BR>"
		End If

		report.Write "Version: " & gpoList(guid)("versionNumber")
		' GPO version numbers are stored in a DWORD with the user settings version
		' number being stored in the high WORD and the computer settings version
		' number being stored in the low WORD.
		report.WriteLine "&nbsp;(Computer: " & (gpoList(guid)("versionNumber") Mod 65536) _
			& ", User: " & (gpoList(guid)("versionNumber") \ 65536) & ")<BR>"

		report.Write "Status of the objects: "
		Select Case gpoList(guid)("flags")
			' "flags" is a bitmap:
			' bit #0 = 0: user settings are enabled
			' bit #0 = 1: user settings are disabled
			' bit #1 = 0: computer settings are enabled
			' bit #1 = 1: computer settings are disabled
			Case 0 report.Write "Enabled"
			Case 1 report.Write "User settings disabled"
			Case 2 report.Write "Computer settings disabled"
			Case 3 report.Write "All settings disabled"
			Case Else report.Write "unknown"
		End Select

		If cfg("includeCounts") Then report.Write "<BR>" & vbNewLine _
			& "Links in the documented Structure: " _
			& (UBound(gpoList(guid)("linksDomain")) _
			+ UBound(gpoList(guid)("linksSite")) + 2)
		If UBound(gpoList(guid)("linksDomain")) >= 0 Then
			report.WriteLine "<BR>" & vbNewLine _
				& "<em>Links in the Domain:</em>" & vbNewLine _
				& "<UL class=""policy"">"
			For Each link In gpoList(guid)("linksDomain")
				report.WriteLine "<LI>" & link & "</LI>"
			Next
			report.WriteLine "</UL>"
		End If
		If UBound(gpoList(guid)("linksSite")) >= 0 Then
			report.WriteLine "<BR>" & vbNewLine _
				& "<em>Links to AD sites:</em>" & vbNewLine _
				& "<UL class=""policy"">"
			For Each link In gpoList(guid)("linksSite")
				report.WriteLine "<LI>" & link & "</LI>"
			Next
			report.WriteLine "</UL>"
		End If

		report.WriteLine "</P>"
	Next

	If cfg("includeCounts") Then report.WriteLine "<P>Group Policy objects total: " & gpoList.Count & "</P>"
End Sub

' ------------------------------------------------------------------------------
' Write the given PSO data to the report.
' ------------------------------------------------------------------------------
Sub WritePSOs(report, PSOArray)
	Dim i
	Dim PSOAppliesTo
	
	PrintDebug "Getting PSOs ..."
	' check if we have any data
	' there could be none if no PSO is present or if permissions are insufficient
	If IsEmpty(PSOArray) then Exit Sub
	For i = 0 To UBound(PSOArray)
		PrintDebug vbTab & "PSO: " & PSOArray(i)("name")
		report.WriteLine "<P class=""pso"">" & symbols("PSO") & "<strong>" _
			& PSOArray(i)("name") & "</strong><BR>"
		report.WriteLine "<TABLE style=""margin-left:20px;"">"
		report.WriteLine "<TR><TD>Minimum Password length</TD><TD><em> " & PSOArray(i)("msDS-MinimumPasswordLength") & " Characters</em></TD></TR>"
		report.WriteLine "<TR><TD>Maximum password age</TD><TD><em> " & PSOArray(i)("msDS-MaximumPasswordAge") & " Day</em></TD></TR>"
		report.WriteLine "<TR><TD>Password History</TD><TD><em> " & PSOArray(i)("msDS-PasswordHistoryLength") & " old Passwordsr</em></TD></TR>"
		report.WriteLine "<TR><TD>Password complexity</TD><TD><em> " & PSOArray(i)("msDS-PasswordComplexityEnabled") & "</em></TD></TR>"
		report.WriteLine "<TR><TD>Reversible encryption</TD><TD><em> " & PSOArray(i)("msDS-PasswordReversibleEncryptionEnabled") & "</em></TD></TR>"
		report.WriteLine "<TR><TD>Account lockout duration</TD><TD><em> " & PSOArray(i)("msDS-lockoutDuration") & " Minutes</em></TD></TR>"
		report.WriteLine "<TR><TD>Account lockout after</TD><TD><em> " & PSOArray(i)("msDS-lockoutThreshold") & " wrong Passwords</em></TD></TR>"
		report.WriteLine "<TR><TD>Deleted lockout counter after</TD><TD><em> " & PSOArray(i)("msDS-lockoutObservationWindow") & " Minutes</em></TD></TR>"
		report.WriteLine "<TR><TD>PSO-priority</TD><TD><em> " & PSOArray(i)("msDS-PasswordSettingsPrecedence") & "</em></TD></TR></TABLE>"
		If IsArray(PSOArray(i)("msDS-PSOAppliesTo")) Then 
			PSOAppliesTo = Join(PSOArray(i)("msDS-PSOAppliesTo"), "</LI><LI>")
		Else
			PSOAppliesTo = PSOArray(i)("msDS-PSOAppliesTo")
		End If
		report.WriteLine "<div style=""margin-left:20px;""> PSO applies for" & vbNewLine & "<em><UL class=""pso""><LI> " & PSOAppliesTo & "</LI></UL></em></div>"
		report.WriteLine "</P>"
	Next

End Sub


' ------------------------------------------------------------------------------
' Write caption to the report.
' ------------------------------------------------------------------------------
Sub WriteCaption(report)
	Dim objectClass, i
	If showPermissionHint = True Then
		report.WriteLine "<div class=""alertbox"">" & vbNewLine & "<H1>Note about Permissions</H1>"
		report.WriteLine "<P>The evaluation of different objects or attributes, there may " _
			& "be permissions problems. This will cause that information to disabled, locked, or " _
			& "expiring accounts - ggf. but also other information - are not properly represented.</P>" _
			& "<P>To get reliable information about these objects and attributes, perform Jos or JoseExec with an " _
			& "account, (e.g. a domain administrator account) has the full read permission to the relevant objects and attributes. " _
			& "Under Windows Vista/Windows 7 and Windows Server 2008 R2 you should start to Jos " _
			& "or JoseExec from a command prompt, you explicitly started as an administrator.</P>" _
			& "<P>(For more information about the specifics of Vista to Windows Server 2008 R2 in " _
			& "<A HREF=""http://www.microsofttranslator.com/bv.aspx?from=de&to=en&a=http://www.faq-o-matic.net/2008/02/22/benutzerkontensteuerung-uac-richtig-einsetzen/"">this</A> " _
			& "article.)</P>"
		If cfg("showAdvancedPermissionHint") Then
			report.WriteLine "<P>Might lack permissions:</P><UL class=""permission"">"
			For i = 0 To UBound(PermissionReason)
				report.WriteLine "<LI>" & PermissionReason(i) & "</LI>"
			Next
			report.WriteLine "</UL>" 
		End If
		report.WriteLine vbNewLine & "</div>"
	End If

	If cfg("includeCounts") And cfg("includeOUs") Then
		report.WriteLine "<H1 id=""stats"">This reports statistics</H1>"
		report.WriteLine "<P>Note: These statistics included only the objects in the scope of this report."
		If cfg("objectFilter") <> "" Then report.WriteLine "(Flat filter: <em>" & cfg("objectFilter") & "</em>)"
		report.WriteLine "<TABLE>" & vbNewLine & "<TR><TH width=""150"">Object class</TD><TH class=""count"">Anzahl</TD></TR>"
		For Each objectClass In stats.Keys
			If Left(objectClass, 3) <> "CN=" Then
				report.WriteLine "<TR><TD>" & objectClass & "</TD><TD class=""count"">" _
					& CStr(stats(objectClass)) & "</TD></TR>"
			End If
		Next
		report.WriteLine "</TABLE>"
	End If

	report.WriteLine "<H1 id=""caption"">Legende</H1>" & vbNewLine & "<TABLE>"
	report.WriteLine "<TR><TD>" & symbols("Domain") & "</TD><TD><P>Domain</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("PSO") & "</TD><TD><P>Password Settings Object</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("organizationalUnit") & "</TD><TD><P>Organizational Unit (OU)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("GPO") & "</TD><TD><P>GPO Link (Group Policy)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("blocked") & "</TD><TD><P>GPO inheritance blocked</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Exclamation") & "</TD><TD><P>Application of Group Policy is enforced</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("container") & "</TD><TD><P>Folder (Container)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("User") & "</TD><TD><P>User</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Computer") & "</TD><TD><P>Computer</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Group") & "</TD><TD><P>Group</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Contact") & "</TD><TD><P>Contact</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Printer") & "</TD><TD><P>Printer</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("disabled") & "</TD><TD><P>Object is disabled (manually)<BR>Link is disabled in Group Policy</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("locked") & "</TD><TD><P>Object is locked (by the system)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Question") & "</TD><TD><P>Unclear status of the object (probably permissions error)</P></TD></TR>"
	report.WriteLine "<TR><TD>" & symbols("Clock") & "</TD><TD><P>Account expires</P></TD></TR>"
	report.WriteLine "</TABLE>"
End Sub

' ------------------------------------------------------------------------------
' Write advertisements to the report.
' ------------------------------------------------------------------------------
Sub WriteAdv(report)
	report.WriteLine "<HR>" & vbNewLine & "<P id=""adv"">" _
		& symbols("Jos") & " Dokumentiert mit <strong>" & cfg("generator") & "</strong> am " _
		& Date & " um " & Time & ".<BR>" & vbNewLine _
		& "Download unter <A HREF=""" & Homepage & """>" _
		& symbols("FOM") & "</A>.</P>"
End Sub

' ==============================================================================
' Data gathering functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Collect domain information on the given domain.
' ------------------------------------------------------------------------------
Function GetDomainInfo(domain)
	Dim domainInfo, objDomain
	Dim objSchema, objPartitions, objConfiguration, objDirectoryService, objRidManager, objInfrastructure
	Dim DCtoQuery
	Dim objExchangeSchemaVersion, objNetBIOS, objDuration, objlockoutObservationWindow
	Dim lockoutThreshold, lockoutDuration, lockoutObservationWindow, hi, lo
	Dim objMaxPwdAge, objMinPwdAge, minPwdLength, pwdHistoryLength, objPwdProperties
	Dim HasRecycleBin
	PrintDebug "Get information about domain " & domain & " ..."

	Set domainInfo = CreateObject("Scripting.Dictionary")
	Set objDomain = GetObject("LDAP://" & domain)
	Set objConfiguration = GetObject("LDAP://" & rootDSE.Get("configurationNamingContext"))
	Set objSchema = GetObject("LDAP://" & rootDSE.Get("schemaNamingContext"))
	Set objPartitions = GetObject("LDAP://CN=Partitions," & rootDSE.Get("configurationNamingContext"))
	Set objDirectoryService = GetObject("LDAP://CN=Directory Service,CN=Windows NT,CN=Services," & rootDSE.Get("configurationNamingContext"))
	Set objRidManager = GetObject("LDAP://CN=RID Manager$,CN=System," & domain)
	Set objInfrastructure = GetObject("LDAP://CN=Infrastructure," & domain)

	domainInfo.Add "DCtoQuery", rootDSE.dnsHostName
	domainInfo.Add "lastBackup", GetADBackupDate(objDomain)
	domainInfo.Add "fqdn", DN2FQDN(domain)
	domainInfo.Add "description", objDomain.Description
	domainInfo.Add "whenCreated", objDomain.Get("whenCreated")
	domainInfo.Add "whenChanged", objDomain.Get("whenChanged")
	domainInfo.Add "rootDomain", DN2FQDN(rootDSE.get("rootDomainNamingContext"))
	domainInfo.Add "sID", Raw2Str(objDomain.Get("objectSID"))
	domainInfo.Add "allDCs", GetAllDCs(domain)

	' Determine FSMO role owners.
	domainInfo.Add "fsmoSchemaMaster", GetName(objSchema.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoDnsMaster", GetName(objPartitions.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoPdcEmulator", GetName(objDomain.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoRidMaster", GetName(objRidManager.Get("fSMORoleOwner"))
	domainInfo.Add "fsmoInfraMaster", GetName(objInfrastructure.Get("fSMORoleOwner"))

	' Forest mode didn't exist prior to Windows Server 2003, thus checking for
	' errors.
	On Error Resume Next

	' Determine operation modes.
	domainInfo.Add "opModeForest", "operation mode is unknown (possibly Windows 2000 environment)"
	Select Case objPartitions.Get("msDS-Behavior-Version")
		Case "0" domainInfo("opModeForest") = "Windows 2000 native"
		Case "1" domainInfo("opModeForest") = "Windows Server 2003 interim"
		Case "2" domainInfo("opModeForest") = "Windows Server 2003 native"
		Case "3" domainInfo("opModeForest") = "Windows Server 2008"
		Case "4" domainInfo("opModeForest") = "Windows Server 2008 R2"
		Case Else TrapError "GetDomainInfo() - forest operation mode"
	End Select
	' the attribute "msDS-Behavior-Version" does not exist in Windows 2000 - clear errors related to this
	if err.number <> 0 then Err.Clear

	domainInfo.Add "opModeDomain", "operation mode is unknown (possibly Windows 2000 environment)"
	Select Case objDomain.Get("msDS-Behavior-Version") 
		Case "0" domainInfo("opModeDomain") = "Windows 2000"
		Case "1" domainInfo("opModeDomain") = "Windows Server 2003"
		Case "2" domainInfo("opModeDomain") = "Windows Server 2003"
		Case "3" domainInfo("opModeDomain") = "Windows Server 2008"
		Case "4" domainInfo("opModeDomain") = "Windows Server 2008 R2"
		Case Else TrapError "GetDomainInfo() - domain operation mode"
	End Select
	' the attribute "msDS-Behavior-Version" does not exist in Windows 2000 - clear errors related to this
	if err.number <> 0 then Err.Clear
	If objDomain.Get("ntMixedDomain") = "1" Then
		domainInfo("opModeDomain") = domainInfo("opModeDomain") & " (mixed)"
	Else
		domainInfo("opModeDomain") = domainInfo("opModeDomain") & " (native)"
	End If

	' Determine schema versions
	domainInfo.Add "schemaVersion", objSchema.Get("objectVersion")
	TrapError "GetDomainInfo() - AD schema version"
	Select Case domainInfo("schemaVersion")
		Case "13" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows 2000 Server)"
		Case "30" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2003)"
		Case "31" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2003 R2)"
		Case "44" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2008)"
		Case "46" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2008 R2 Beta/RC)"
		Case "47" domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (Windows Server 2008 R2)"
		Case Else domainInfo("schemaVersion") = domainInfo("schemaVersion") & " (unbekannte Version!)"
	End Select

	Set objExchangeSchemaVersion = GetObject("LDAP://CN=ms-Exch-Schema-Version-Pt," & objSchema.Get("distinguishedName"))
	If objExchangeSchemaVersion Is Nothing Then
		Err.Clear
	Else
		domainInfo.Add "exchSchemaVersion", objExchangeSchemaVersion.Get("rangeUpper")
		Dim exchangeSchemaVersion : exchangeSchemaVersion = objExchangeSchemaVersion.Get("rangeUpper")
		TrapError "GetDomainInfo() - Exchange schema version"
		If exchangeSchemaVersion <> "" Then
			Select Case domainInfo("exchSchemaVersion")
				Case "4397" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange 2000 Server)"
				Case "4406" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange 2000 Server SP3)"
				Case "6870" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2003)"
				Case "6936" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2003 SP3)"
				Case "10637" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2007 RTM)"
				Case "11116" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2007 SP1)"
				Case "14622" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2010 RTM oder Exchange Server 2007 SP2)"
				Case "14726" domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (Exchange Server 2010 SP1)"
				Case Else domainInfo("exchSchemaVersion") = domainInfo("exchSchemaVersion") & " (unbekannte Version!)"
			End Select
		End If
		Set objExchangeSchemaVersion = Nothing
	End If

	' NetBIOS name is read from the crossRef object in the partition container.
	Set objNetBIOS = QueryAD("SELECT nETBIOSName FROM '" & objPartitions.aDSPath _
		& "' " & "WHERE nCName = '" & domain & "'")
	domainInfo.Add "netBiosName", UCase(objNetBIOS("nETBIOSName"))
	TrapError "GetDomainInfo() - NetBIOS name"
	Set objNetBIOS = Nothing

	On Error Goto 0

	' *** Get settings for deleted objects
	' settings are empty by default, so suppress errors
	PrintDebug "Getting settings for deleted objects ..."
	On Error Resume Next
	domainInfo.Add "tombstoneLifetime", 0
	domainInfo("tombstoneLifetime") = objDirectoryService.Get("tombstoneLifetime")
	domainInfo.Add "msDS-DeletedObjectLifetime", 0
	domainInfo("msDS-DeletedObjectLifetime") = objDirectoryService.Get("msDS-DeletedObjectLifetime")
	On Error Goto 0

	' *** Query Recycle Bin Feature
	domainInfo.Add "HasRecycleBin", False
	On Error Resume Next
	' this works only as long as there is only the Recycle Bin optional feature ...
	' in the future we need to make this more intelligent!
	HasRecycleBin = objPartitions.Get("msDS-EnabledFeature")
	On Error Goto 0

	domainInfo("HasRecycleBin") = HasRecycleBin <> ""
	
	' *** Read Password Policy Settings ***
	' it seems safe to read these settings from the domain object as there apparently
	' is a process that takes the effective settings from the GPOs and writes them to
	' the domain head. Moreover, there is even code that writes changes to the domain head
	' back to the Default Domain Policy (no matter if the effective settings come from there ...)
	' compare: http://x220.minasi.com/forum/topic.asp?whichpage=1&TOPIC_ID=20647&#93786
	PrintDebug "Getting password settings ..."
	' Get account lockout settings.
	domainInfo.Add "lockoutThreshold", objDomain.Get("lockoutThreshold")

	domainInfo.Add "lockoutDuration", 0
	' there might be problems with lockoutDuration, so we'll catch errors
	' as we need the value later on
	' (well, at least I ran across an error once ...)
	Set objDuration = objDomain.lockoutDuration
	domainInfo("lockoutDuration") = Integer8Number(objDuration)
	If IsEmpty(domainInfo("lockoutDuration")) Then
	' make sure we don't get a Null value
		domainInfo("lockoutDuration") = -1
		PrintDebug "Error retrieving lockoutDuration; setting value to -1"
		showPermissionHint = True
		SetPermissionReason "Sperrdauer Domne" 
	End If

	' Get account lockout observation settings.
	domainInfo.Add "lockoutObservationWindow", 0
	Set objlockoutObservationWindow = objDomain.lockoutObservationWindow
	domainInfo("lockoutObservationWindow") = Integer8Number(objlockoutObservationWindow)

	' Get minimum password age
	domainInfo.Add "minPwdAge", 0
	Set objMinPwdAge = objDomain.minPwdAge
	domainInfo("minPwdAge") = Integer8Number(objMinPwdAge) / 60 / 24
	
	' Get maximum password age
	domainInfo.Add "maxPwdAge", 0
	Set objMaxPwdAge = objDomain.maxPwdAge
	domainInfo("maxPwdAge") = Integer8Number(objMaxPwdAge) / 60 / 24
	
	' Get password restrictions
	domainInfo.Add("minPwdLength"), objDomain.minPwdLength
	domainInfo.Add("pwdHistoryLength"), objDomain.pwdHistoryLength
	domainInfo.Add("pwdProperties"), objDomain.pwdProperties
	
	' ### Count data in the root container unless an object filter is specified
	' hm ... seems we do not need that here
	'PrintDebug "objectFilter is '" & cfg("objectFilter") & "'"
	'If cfg("objectFilter") = "" then GetCount(objDomain)
	
	Set objInfrastructure = Nothing
	Set objRidManager = Nothing
	Set objPartitions = Nothing
	Set objSchema = Nothing
	Set objDomain = Nothing

	Set GetDomainInfo = domainInfo
End Function

' ------------------------------------------------------------------------------
' Get all DCs in the given domain.
' See:
' - http://www.selfadsi.de/extended-ad/search-domain-controllers.htm
' ------------------------------------------------------------------------------
Function GetAllDCs(domain)
	Dim ldapFilter, objectList, dcDN, dcDNS
	Dim ado
	Dim arrDCs, i
	
	printDebug "Getting all DCs for domain " & domain & " ..."
	ldapFilter = "(&(objectCategory=computer)(userAccountControl:1.2.840.113556.1.4.803:=8192))"

	Set ado = CreateObject("ADODB.Connection")
	ado.Provider = "ADSDSOObject"
	ado.Open "ADSearch" 
	Set objectList = ado.Execute("<LDAP://" & domain & ">;" & ldapFilter & ";distinguishedName,dnsHostName;subtree")

	arrDCs = Array()
	i=0
	While Not objectList.EOF
		ReDim preserve arrDCs(i)
		dcDN = objectList.Fields("distinguishedName")
		dcDNS = objectList.Fields("dnsHostName")
		arrDCs(i) = dcDNS & " (" & dcDN & ")"
		PrintDebug "... " & arrDCs(i)
		objectList.MoveNext
		i = i + 1
	Wend 
	GetAllDCs = arrDCs
End Function

' ------------------------------------------------------------------------------
' Get trusts for the given domain.
' Trust data structures are documented here:
' - http://msdn2.microsoft.com/en-us/library/aa392340.aspx
' - http://msdn2.microsoft.com/en-us/library/ms722477.aspx
' ------------------------------------------------------------------------------
Function GetTrusts(domain)
	Dim attributes, wql, objTrust, arrTrusts, i, TrustAttr, TrustAttrString

	PrintDebug "Get trusts for domain " & domain & " ..."

	GetTrusts = Array()

	attributes = "name,flatName,trustAttributes,trustDirection,trustType"
	wql = "SELECT " & attributes & " FROM 'LDAP://CN=System," & domain _
		& "' WHERE objectClass='trustedDomain'"
	On Error Resume Next
	Set objTrust = QueryAD(wql)
	If TrapError("GetTrusts(" & wql & ")") Then 
		showPermissionHint = True
		SetPermissionReason "Trust Relationships"
		Exit Function
	End If
	On Error Goto 0

	If Not objTrust.EOF Then objTrust.MoveFirst
	arrTrusts = Array()
	i = 0
	Do Until objTrust.EOF
		ReDim Preserve arrTrusts(i)
		Set arrTrusts(i) = CreateObject("Scripting.Dictionary")
		arrTrusts(i).Add "name", objTrust("name").value
		PrintDebug "   Adding trust " & objTrust("name")
		arrTrusts(i).Add "flatName", objTrust("flatName").value

		arrTrusts(i).Add "direction", Null
		Select Case objTrust("trustDirection")
			Case 1 arrTrusts(i)("direction") = "inbound trust relationship"
			Case 2 arrTrusts(i)("direction") = "outbound trust relationship"
			Case 3 arrTrusts(i)("direction") = "bi-directional trust relationship"
		End Select

		arrTrusts(i).Add "type", "unknown relationship type"
		Select Case objTrust("trustType")
			Case 1 arrTrusts(i)("type") = "Relationship to a Windows NT 4.0 Domain"
			Case 2 arrTrusts(i)("type") = "Relationship to an Active Directory Domain (Windows 2000 or later)"
			Case 3 arrTrusts(i)("type") = "Relationship to a Kerberos Realm"
			Case 4 arrTrusts(i)("type") = "Relationship to a DCE-Realm"
		End Select
		
		arrTrusts(i).Add "attributes", Null
		TrustAttr = objTrust("trustAttributes").value
		arrTrusts(i)("attributes") = DecodeTrustFlags(TrustAttr) & " (" & TrustAttr & ")"

		i = i + 1
		objTrust.MoveNext
	Loop
	GetTrusts = arrTrusts
End Function

' ------------------------------------------------------------------------------
' Return a dictionary with the OUs of the given domain. The dictionary has the
' following structure:
'
' "class"       [string]     object class (can be either "builtinDomain",
'                            "container" or "organizationalUnit")
' "name"        [string]     the object's name
' "level"       [integer]    the object's level in the OU hierarchy
' "description" [string]     descriptive comment
' "whenCreated" [timestamp]  the object's creation time
' "whenChanged" [timestamp]  time when the object was last changed
' "count"       [dictionary] (optional) total and per-category counts of
'                            contained objects (users, computers, printers, ...)
' "gpoInherit"  [boolean]    are GPOs inherited from parent containers?
' "GPO"         [array]      GPOs assigned to the OU
' "User"        [array]      users assigned to the OU
' "Computer"    [array]      computers assigned to the OU
' "Group"       [array]      groups assigned to the OU
' "Contact"     [array]      contacts assigned to the OU
' "Printer"     [array]      printers assigned to the OU
' "children"    [array]      (optional) array of dictionaries of child objects
'                            of the same class
' ------------------------------------------------------------------------------
Function GetOUs(ouClass, dn, level)
	Dim objStart, objSub, thisOU, options

	PrintDebug ouClass & ": " & dn

	' catch error in case of missing permissions
	On Error Resume Next
	Set objStart = GetObject("LDAP://" & EscapeDN(dn))
	If TrapError("Error binding to object!") Then
		showPermissionHint = True
		SetPermissionReason "OU-Zugriff " & dn
		Exit Function
	End If
	On Error Goto 0

	Set thisOU = CreateObject("Scripting.Dictionary")

	thisOU.Add "class", ouClass
	thisOU.Add "name", objStart.Get("name")
	thisOU.Add "level", level
	thisOU.Add "description", ""
	' Ignore error in case "description" is empty.
	On Error Resume Next
	thisOU("description") = objStart.Get("description")
	On Error Goto 0
	thisOU.Add "whenCreated", objStart.Get("whenCreated")
	thisOU.Add "whenChanged", objStart.Get("whenChanged")
	thisOU.Add "gpoInherit", True
	On Error Resume Next
	options = &h00000000
	options = objStart.Get("gpOptions")
	thisOU("gpoInherit") = Not (CInt(options) And GPO_IGNORE)
	On Error Goto 0

	If InStr(1, dn, cfg("objectFilter"), vbTextCompare) Then
		If ouClass = "organizationalUnit" Or level > 0 Then
			' Get data of non-container objects if requested
			If cfg("includeCounts") Then thisOU.Add "count", GetCount(objStart)
			If cfg("includeGPOs") Then thisOU.Add "GPO", GetLinkedGPOs(objStart)
			If cfg("includeUsers") Then thisOU.Add "User", GetObjects("user", objStart, arrUserAttr)
			If cfg("includeComputers") Then thisOU.Add "Computer", GetObjects("computer", objStart, arrComputerAttr)
			If cfg("includeGroups") Then thisOU.Add "Group", GetObjects("group", objStart, arrGroupAttr)
			If cfg("includeContacts") Then thisOU.Add "Contact", GetObjects("contact", objStart, arrContactAttr)
			If cfg("includePrinters") Then thisOU.Add "Printer", GetObjects("printer", objStart, arrPrinterAttr)
		End If
	End If

	' Recurse only for objects of the same ouClass.
	objStart.Filter = Array(ouClass)
	If IsMatching(dn, cfg("objectFilter")) Then
		Dim i : i = 0
		Dim children : children = Array()
		For Each objSub In objStart
			If IsMatching(objSub.Get("distinguishedName"), cfg("objectFilter")) Then
				Dim name : name = objSub.Get("name")
				' Skip system containers, foreign security principals, Exchange
				' system objects and program data containers.
				Select Case LCase(name)
					Case "system"
					Case "foreignsecurityprincipals"
					Case "microsoft exchange system objects"
					Case "program data"
					Case Else
						ReDim Preserve children(i)
						'Set children(i) = GetOUs(ouClass, Replace(objSub.Get("distinguishedName"), "/", "\/"), level + 1)
						Set children(i) = GetOUs(ouClass, objSub.Get("distinguishedName"), level + 1)
						i = i + 1
				End Select
			End If
		Next
		thisOU.Add "children", children
	End If

	Set objStart = Nothing

	Set GetOUs = thisOU
End Function

' ------------------------------------------------------------------------------
' Return a dictionary with the number of objects in the given container (total
' as well as by category).
' ------------------------------------------------------------------------------
Function GetCount(objStart)
	Dim count, obj, objectClass

	Set GetCount = Nothing

	' this line prevented the function from counting objects in root
	' seems we should delete it
	'###If InStr(Join(objStart.Get("objectClass")), "domain") > 0 Then Exit Function

	' PrintDebug "Get counts for '" & objStart.Get("name") & "' ..."

	Set GetCount = CreateObject("Scripting.Dictionary")

	count = 0
	For Each obj In objStart
		count = count + 1
	Next
	GetCount.Add "total", count

	For Each objectClass In Array("user", "computer", "group", "contact", "printQueue", "organizationalUnit")
		objStart.Filter = Array(objectClass)
		count = 0
		For Each obj In objStart
			' Increment count unless a user actually is a computer.
			If Not (objectClass = "user" And IsComputer(obj.Get("objectCategory"))) Then count = count + 1
		Next
		GetCount.Add objectClass, count
		' Count objects for (global) statistics.
		stats(objectClass) = stats(objectClass) + count
		' PrintDebug "Count of " & objectClass & ": " & stats(objectClass)
	Next
End Function

' ------------------------------------------------------------------------------
' Return an array of dictionaries representing the GPOs linked to the given
' object. Each dictionary contains the following information:
'
' "guid"        [string]  the GPO's GUID
' "disabled"    [boolean] GPO is enabled/disabled for the object
' "enforced"    [boolean] GPO is or isn't enforced for the object
' ------------------------------------------------------------------------------
Function GetLinkedGPOs(objStart)
	Dim gpLinks, re, i, link, linkStatus

	Dim arrGPOs : arrGPOs = Array()

	gpLinks = ""
	On Error Resume Next
	gpLinks = Trim(objStart.Get("gPLink"))
	On Error Goto 0

	If gpLinks <> "" Then
		Set re = New RegExp
		re.Pattern = ".*?{(.+?)}.*?;(.+)"
		re.Global = True
		re.IgnoreCase = True

		If Left(gpLinks, 1) = "[" Then gpLinks = Mid(gpLinks, 2)
		If Right(gpLinks, 1) = "]" Then gpLinks = Mid(gpLinks, 1, InStrRev(gpLinks, "]") - 1)
		i = 0
		For Each link In Split(gpLinks, "][")
			PrintDebug vbTab & "Linked GPO: " & link
			If re.Test(link) Then
				ReDim Preserve arrGPOs(i)
				Set arrGPOs(i) = CreateObject("Scripting.Dictionary")

				Dim matches : Set matches = re.Execute(link)
				arrGPOs(i).Add "guid", matches(0).SubMatches(0)
				arrGPOs(i).Add "disabled", False
				arrGPOs(i).Add "enforced", False

				' "Enforce" bit MUST be ignored when "ignore" bit is set.
				' See: <http://msdn.microsoft.com/en-us/library/cc232505.aspx>
				linkStatus = CInt(matches(0).SubMatches(1))
				If linkStatus And GPO_IGNORE Then
					arrGPOs(i)("disabled") = True
				ElseIf linkStatus And GPO_ENFORCE Then
					arrGPOs(i)("enforced") = True
				End If
				Set matches = Nothing

				i = i + 1
			End If
		Next
	End If

	GetLinkedGPOs = arrGPOs
End Function

' ------------------------------------------------------------------------------
' Return an array of dictionaries representing the objects of the given
' category that are contained by objStart. Each dictionary contains the
' following elements:
'
' "name"        [string]     the object's name
' "description" [string]     descriptive comment
' "disabled"    [boolean]    (optional) is the object disabled?
' "expires"     [timestamp]  the object's expiry date (only category "user")
' "locked"      [boolean]    is the object locked?
' "lockoutTime" [timestamp]  (optional) the time the object was locked
' "attributes"  [dictionary] selected attributes of the object
' "Printer"     [array]      (only category "computer") print queues shared by
'                            the computer
' ------------------------------------------------------------------------------
Function GetObjects(category, objStart, attributes)
	Dim objSub, i

	PrintDebug "Get " & category & " objects in '" & objStart.Get("name") & "' ..."

	Dim arrObjects : arrObjects = Array()

	If category = "printer" Then
		objStart.Filter = Array("printQueue")
	Else
		objStart.Filter = Array(category)
	End If

	i = 0
	For Each objSub In objStart
		If category <> "user" Or Not IsComputer(objSub.Get("objectCategory")) Then
			PrintDebug vbTab & "Add " & objSub.Get("name")
			ReDim Preserve arrObjects(i)
			Set arrObjects(i) = CreateObject("Scripting.Dictionary")
			arrObjects(i).Add "name", objSub.Get("name")
			arrObjects(i).Add "description", ""
			On Error Resume Next
			arrObjects(i)("description") = objSub.Get("description")
			On Error Goto 0
			If category = "user" Or category = "computer" Then 
				arrObjects(i).Add "disabled", IsDisabled(objSub)
				arrObjects(i).Add "locked", IsLocked(objSub)
			End If	
			If category = "user" Then arrObjects(i).Add "expires", GetExpiryDate(objSub)
			If arrObjects(i)("locked") Then arrObjects(i).Add "lockoutTime", GetLockoutTime(objSub)
			arrObjects(i).Add "attributes", GetAttributes(objSub, attributes)
			If category = "computer" Then arrObjects(i).Add "Printer", GetObjects("printer", objSub, arrPrinterAttr)
			i = i + 1
		End If
	Next

	GetObjects = arrObjects
End Function

' ------------------------------------------------------------------------------
' Return an array with the specified attributes of the given object.
' ------------------------------------------------------------------------------
Function GetAttributes(objSub, attributes)
	Dim i, name, attribs

	PrintDebug vbTab & "Get attributes of " & objSub.Get("name") & " ..."

	Set attribs = CreateObject("Scripting.Dictionary")

	For i = LBound(attributes) To UBound(attributes)
		name = Trim(attributes(i))
		attribs.Add name, ""

		On Error Resume Next
		attribs(name) = objSub.Get(name)
		If Err.Number <> 0 Then
			' In case the attribute needs to be requested first.
			objSub.GetInfoEx Array(name), 0
			attribs(name) = objSub.Get(name)
		End If
		On Error Goto 0

		Select Case LCase(name)
			Case "" attribs.Remove(name)
			Case "objectsid"
				attribs(name) = Raw2Str(objSub.ObjectSid)
			Case "sidhistory"
				' sIDHistory is stored as an array.
				If IsArray(attribs(name)) Then
					If LCase(TypeName(attribs(name))) = "variant()" Then
						Dim j
						For j = LBound(attribs(name)) To UBound(attribs(name))
							attribs(name)(j) = Raw2Str(attribs(name)(j))
						Next
					Else
						attribs(name) = Raw2Str(attribs(name))
					End If
				End If
			Case "streetaddress"
				Dim postcode, city
				On Error Resume Next
				postcode = objSub.Get("postalCode")
				city = objSub.Get("l")
				On Error Goto 0
				If attribs(name) <> "" And (postcode <> "" Or city <> "") Then
					attribs(name) = Trim(attribs(name) & ", " & Trim(postcode & " " & city))
				Else
					attribs(name) = Trim(attribs(name) & " " & Trim(postcode & " " & city))
				End If
			Case "grouptype"
				Dim group
				If attribs(name) < 0 Then
					group = " (Security Group)"
					attribs(name) = attribs(name) + 2147483648
				Else
					group = " (Distribution Group)"
				End If
				Select Case attribs(name)
					Case 2 attribs(name) = "Global Group"
					Case 4 attribs(name) = "Domain local Group"
					Case 5 attribs(name) = "Built-in Group"
					Case 8 attribs(name) = "Universal Group"
					Case Else attribs(name) = "Unknown (" & attribs(name) & ")"
				End Select
				attribs(name) = attribs(name) & group
			Case "dcroles"
				attribs(name) = GetDCRoles(objSub)
			Case "lastlogontimestamp"
				' this exists only in forest mode 2003 or above
				' so we need to handle errors
				On Error Resume Next
				attribs(name) = Integer8Date(objSub.Get(name), systemBias)
				If Err.number <> 0 Then
					attribs(name) = ""
					Err.Clear
				End If
				On Error Goto 0
			Case "logonhours"
				attribs(name) = GetLogonHours(objSub)
			Case "pwdlastset"
				On Error Resume Next
				attribs(name) = Integer8Date(objSub.Get(name), systemBias)
				if attribs(name) = #1/1/1601# then attribs(name) = "niemals"
				On Error Goto 0
			Case "useraccountcontrol"
				Dim flags : flags = 0
				On Error Resume Next
				flags = objSub.UserAccountControl
				On Error Goto 0
				If flags <> 0 Then
					attribs(name) = CStr(flags) & ": " & DecodeUACFlags(flags)
				Else
					attribs(name) = ""
				End If
			Case "terminalserviceshomedirectory"
				' this exists only in Windows Server 2003 and later
				On Error Resume Next
				attribs(name) = objSub.TerminalServicesHomeDirectory
				On Error Goto 0
			Case "terminalservicesprofilepath"
				' this exists only in Windows Server 2003 and later
				On Error Resume Next
				attribs(name) = objSub.TerminalServicesProfilePath
				On Error Goto 0
		End Select
		If IsArray(attribs(name)) Then
			PrintDebug vbTab & vbTab & name & " = " & Join(attribs(name), ",")
		Else
			PrintDebug vbTab & vbTab & name & " = " & attribs(name)
		End If
	Next

	Set GetAttributes = attribs
End Function

' ------------------------------------------------------------------------------
' Return a dictionary of dictionaries. The parent dictionary contains child
' dictionaries representing the GPOs defined in the given domain, which are
' identified by the GPO's respective GUID. Each child dictionary contains the
' following elements:
'
' "displayName"   [string]    the name the GPO is displayed as in the GUI
' "whenCreated"   [timestamp] the GPO's creation time
' "whenChanged"   [timestamp] time when the GPO was last changed
' "versionNumber" [DWORD]     the GPO's version number
' "flags"         [integer]   bitmap defining if computer/user settings are
'                             enabled or disabled
' "comment"       [string]    descriptive comment
' "gpcWMIFilter"  [string]    (optional) name of the WMI filter associated with
'                             the GPO
' "linksDomain"   [array]     locations where the GPO is linked in the domain
' "linksSite"     [array]     locations where the GPO is linked in the site
'
' This is a dictionary of dictionaries instead of an array of dictionaries,
' because when writing the information which GPO is linked to a particular
' container, properties of the respective GPO are read from this collection.
' Accessing a dictionary for that matter is far more efficient than accessing
' an array.
' ------------------------------------------------------------------------------
Function GetGPOs(domain)
	Dim attributes, query, objGPO, theGPOs, guid

	PrintDebug "Get group policy objects ..."

	attributes = "aDSPath,displayName,name,whenCreated,whenChanged,versionNumber" _
		& ",flags,gPCFileSysPath"
	query = "SELECT " & attributes & " FROM 'LDAP://CN=Policies,CN=System," & domain _
		& "' WHERE objectClass = 'groupPolicyContainer' ORDER BY displayName"

	On Error Resume Next
	Set objGPO = QueryAD(query)
	If TrapError("GetGPOs()") Or objGPO Is Nothing Then Exit Function

	Set theGPOs = CreateObject("Scripting.Dictionary")

	Do Until objGPO.EOF
		guid = Replace(Replace(objGPO("name").Value, "{", ""), "}", "")
		PrintDebug vbTab & "GPO: " & guid
		theGPOs.Add guid, CreateObject("Scripting.Dictionary")

		theGPOs(guid).Add "displayName", objGPO("displayName").Value
		theGPOs(guid).Add "whenCreated", objGPO("whenCreated").Value
		theGPOs(guid).Add "whenChanged", objGPO("whenChanged").Value
		theGPOs(guid).Add "versionNumber", objGPO("versionNumber").Value
		theGPOs(guid).Add "flags", objGPO("flags").Value
		theGPOs(guid).Add "comment", GetGPOComment(objGPO("gPCFileSysPath").Value)
		theGPOs(guid).Add "gpcWMIFilter", GetWMIFilter(objGPO("adsPath").Value)

		If Err.Number <> 0 Then Err.Clear

		theGPOs(guid).Add "linksDomain", GetGPOLinks(objGPO("name").Value, domain)
		theGPOs(guid).Add "linksSite", GetGPOLinks(objGPO("name").Value, rootDSE.Get("configurationNamingContext"))

		objGPO.MoveNext
	Loop
	On Error Goto 0

	Set GetGPOs = theGPOs
End Function

' ------------------------------------------------------------------------------
' Return data on the last AD backup: date, time, and performing DC
' This data is stored in two attributes, one of which is XML-formatted - 
' we'll read this attribute
' ------------------------------------------------------------------------------
Function GetADBackupDate(objDom)
	Dim ReplValue, Subvalue, DSASignature
	Dim timeLeft, timeRight, DCLeft, DCRight
	Dim TimeVal, DCName, BackupDate, BackupTime
	
	' the attribute was introduced in Windows 2003, so we'll catch errors
	On Error Resume Next
	objDom.GetInfoEx Array("msDS-ReplAttributeMetaData"),0
	ReplValue = objDom.Get("msDS-ReplAttributeMetaData")
	If Err.number <> 0 Then 
		GetADBackupDate = "(No information available)"
		Exit Function
	End If
	On Error Goto 0

	DSASignature = ""
	For Each Subvalue In ReplValue
		if instr(LCase(SubValue), "dsasignature") >0 then DSASignature = Subvalue
	Next
	' this is not too pretty but after all we do not need an XML parser here
	If DSASignature <> "" Then
		timeLeft = InStr(LCase(DSASignature), LCase("<ftimeLastOriginatingChange>")) + Len("<ftimeLastOriginatingChange>")
		timeRight = InStr(timeLeft, LCase(DSASignature), LCase("</ftimeLastOriginatingChange>"))
		DCLeft = InStr(LCase(DSASignature), LCase("<pszLastOriginatingDsaDN>")) + Len("<pszLastOriginatingDsaDN>") + Len("CN=NTDS Settings,CN=")
		DCRight = InStr(DCLeft, DSASignature, ",")
		TimeVal = Mid(DSASignature, timeLeft, timeRight - timeLeft)
		On Error Resume Next
		' DC Name might be missing
		DCName = Mid(DSASignature, DCLeft , DCRight - DCLeft)
		if err.number <> 0 then DCName = "(kein Wert)"
		On Error Goto 0
		BackupDate = Left(TimeVal, InStr(TimeVal, "T") - 1)
		BackupTime = Mid(TimeVal, InStr(TimeVal, "T") + 1)
		BackupTime = Left(BackupTime, Len(BackupTime) - 1)
		BackupTime = CDate(BackupDate & " " & BackupTime)
		BackupTime = DateAdd("n", -(systemBias), BackupTime)
		GetADBackupDate = BackupTime & " von " & DCName
	Else
		' AD seems not to have been backed up yet
		GetADBackupDate = "so far not backed up"
	End If
	
End Function

' ------------------------------------------------------------------------------
' Return an array of dictionaries with the Password Settings Objects defined In the domain.
' The array of dictionaries contains the following data:
' - single values for: name, msDS-MaximumPasswordAge, msDS-MinimumPasswordAge,
'   msDS-MinimumPasswordLength, msDS-PasswordHistoryLength, msDS-PasswordComplexityEnabled, 
'   msDS-PasswordReversibleEncryptionEnabled, msDS-lockoutObservationWindow, 
'   msDS-lockoutDuration, msDS-lockoutThreshold, msDS-PasswordSettingsPrecedence
' - multiple values (array) for: msDS-PSOAppliesTo
' ------------------------------------------------------------------------------
Function GetPSOs(domain)
	Dim PSOContainer, PSO, thePSOs, i
	
	' this requires admin rights, and of course the feature might be absent, so catch errors
	On Error Resume Next
	Set PSOContainer = GetObject("LDAP://CN=Password Settings Container,CN=System," & domain)
	If Err.number <> 0 Then
		showPermissionHint = True
		SetPermissionReason "Access PSO"
		Exit Function
	End If
	On Error Goto 0
	
	PSOContainer.Filter = Array("msDS-PasswordSettings")
	thePSOs = Array()
	i = 0
	For Each PSO In PSOContainer
		ReDim Preserve thePSOs(i)
		Set thePSOs(i) = CreateObject("Scripting.Dictionary")
		thePSOs(i).Add "name", PSO.Get("name")
		thePSOs(i).Add "msDS-MinimumPasswordLength", PSO.Get("msDS-MinimumPasswordLength")
		thePSOs(i).Add "msDS-PasswordHistoryLength", PSO.Get("msDS-PasswordHistoryLength")
		thePSOs(i).Add "msDS-PasswordComplexityEnabled", PSO.Get("msDS-PasswordComplexityEnabled")
		thePSOs(i).Add "msDS-PasswordReversibleEncryptionEnabled", PSO.Get("msDS-PasswordReversibleEncryptionEnabled")
		thePSOs(i).Add "msDS-lockoutThreshold", PSO.Get("msDS-lockoutThreshold")
		thePSOs(i).Add "msDS-PasswordSettingsPrecedence", PSO.Get("msDS-PasswordSettingsPrecedence")
		thePSOs(i).Add "msDS-PSOAppliesTo", PSO.Get("msDS-PSOAppliesTo")
		thePSOs(i).Add "msDS-MaximumPasswordAge", Integer8Number(PSO.Get("msDS-MaximumPasswordAge"))/60/24
		thePSOs(i).Add "msDS-MinimumPasswordAge", Integer8Number(PSO.Get("msDS-MinimumPasswordAge"))/60/24
		thePSOs(i).Add "msDS-lockoutObservationWindow", Integer8Number(PSO.Get("msDS-lockoutObservationWindow"))
		thePSOs(i).Add "msDS-lockoutDuration", Integer8Number(PSO.Get("msDS-lockoutDuration"))
		i = i + 1
	Next
	GetPSOs = thePSOs
	Set PSOContainer = Nothing

End Function


' ==============================================================================
' Helper functions
' ==============================================================================

' ------------------------------------------------------------------------------
' Create a dictionary for storing attributes that should be read from AD.
' Include attributes "whenCreated" and "whenChanged" by default.
' ------------------------------------------------------------------------------
Function CreateAttributeSet()
	Set CreateAttributeSet = CreateObject("Scripting.Dictionary")
	CreateAttributeSet.Add "whenCreated", True
	CreateAttributeSet.Add "whenChanged", True
End Function

' ------------------------------------------------------------------------------
' Evaluate a definition line. If the definition has the form "key = value" an
' array [key, value] is returned. If the definition is a flag (i.e. just "key"
' without a "value") then an array [key, True] is returned.
' ------------------------------------------------------------------------------
Function EvalDefinition(line)
	Dim def : def = Split(line, "=", 2)
	def(KEY) = Trim(def(KEY))
	If UBound(def) = KEY Then
		ReDim Preserve def(VALUE)
		def(VALUE) = True
	Else
		def(VALUE) = Trim(def(VALUE))
	End If
	EvalDefinition = def
End Function

' ------------------------------------------------------------------------------
' Check whether an account is disabled or not. Return True if the account is
' disabled, False if it's disabled, Null if the attribute can't be read for
' some reason.
' ------------------------------------------------------------------------------
Function IsDisabled(objAccount)
	Dim status : status = 0

	On Error Resume Next
	Err.Clear
	status = objAccount.UserAccountControl
	If Err.Number <> 0 Or status = 0 Then
		' perhaps a permissions error
		showPermissionHint = True
		SetPermissionReason "Aktivierungsstatus " & objAccount.Name
		IsDisabled = Null
	ElseIf (status And ADS_UF_ACCOUNTDISABLE) Then
		IsDisabled = True
	Else
		IsDisabled = False
	End If
	On Error Goto 0
End Function

' ------------------------------------------------------------------------------
' Return the expiry date of the given object or Null if an error occurs.
' ------------------------------------------------------------------------------
Function GetExpiryDate(objSub)
	On Error Resume Next
	Err.Clear
	GetExpiryDate = Integer8Date(objSub.AccountExpires, systemBias)
	If Err.Number <> 0 Then
		PrintDebug "Can't get expiry date for object " & objSub.Get("name")
		' perhaps a permissions error
		showPermissionHint = True
		SetPermissionReason "Expiration time " & objSub.Get("name")
		GetExpiryDate = Null
	End If
	On Error Goto 0
End Function

' ------------------------------------------------------------------------------
' Check whether an account is disabled or not. Return True if the account is
' disabled, False if it's disabled, Null if the attribute can't be read for
' some reason.
' ------------------------------------------------------------------------------
Function IsLocked(objAccount)
	Dim status

	On Error Resume Next
	Err.Clear
	status = objAccount.UserAccountControl
	If Err.Number <> 0 Or status = 0 Then
		' perhaps a permissions error
		showPermissionHint = True
		SetPermissionReason "Lock status " & objAccount.Name
		IsLocked(0) = Null
	ElseIf (status And ADS_UF_LOCKOUT) Then
		IsLocked = True
	lse
		IsLocked = False
	End If
End Function

' ------------------------------------------------------------------------------
' Return the time the given account was locked out or Null in case of an error.
' ------------------------------------------------------------------------------
Function GetLockoutTime(objAccount)
	On Error Resume Next
	Err.Clear
	GetLockoutTime = Integer8Date(objAccount.LockoutTime, systemBias)
	If TrapError("GetLockoutTime()") Then 
		GetLockoutTime = Null
		showPermissionHint = True
		SetPermissionReason "Blocking time for " & objAccount.Name
	End If
	On Error Goto 0
End Function

' ------------------------------------------------------------------------------
' Return the logon hours for the given account.
' Based on <http://www.rlmueller.net/Document%20LogonHours.htm>
' ------------------------------------------------------------------------------
Function GetLogonHours(objAccount)
	Dim DayNames
	Dim AllLogonHours(20)
	Dim LogonHoursBits(167)
	Dim LogonHours
	Dim LogonHourByte, LogonHour
	Dim k, Counter, LoopCounter, j, m
	Dim DayOfWeek(6), DayLogon(6)
	Dim DayCount, HourCount

	' German day names (short), beginning with Sunday (US-like)
	DayNames = Array("So", "Mo", "Di", "Mi", "Do", "Fr", "Sa")
	
	' Retrieve the user's logonHours attribute.
	On Error Resume Next
	objAccount.GetInfoEx Array("logonHours"), 0
	LogonHours = objAccount.Get("logonHours")
	' no logon hours for the account - return
	'   (this happens usually if the object was created via script -
	'   effectively, this means logon is allowed any time. However,
	'   we will not return any value)
	If Err.Number <> 0 Then Exit Function
	On Error Goto 0
	
	' Populate a byte array.
	For k = 1 To LenB(LogonHours)
		AllLogonHours(k - 1) = AscB(MidB(LogonHours, k, 1))
	Next
	
	' Populate a bit array, offset by the time zone bias.
	j = 0
	For Each LogonHourByte In AllLogonHours
		For k = 7 To 0 Step -1
			m = 8*j + k - (systemBias/60)
			If (m < 0) Then
				m = m + 168
			End If
			' *** correction Nils Kaczenski:
			' wrap the other way as well!
			If (m > 167) Then
				m = m - 168
			End If
			If (LogonHourByte And 2^k) Then
				LogonHoursBits(m) = 1
			Else
				LogonHoursBits(m) = 0
			End If
		Next
		j = j + 1
	Next
	
	Counter = 0
	LoopCounter = 0
	DayCount = 0
	
	For Each LogonHour In LogonHoursBits
		' add logon hours to an array of days
		DayOfWeek(DayCount) = DayOfWeek(DayCount) & LogonHour
		Counter  = Counter + 1
		If (Counter = 24) Then
			' this makes a full day, so check it
			If DayOfWeek(DayCount) = String(24, "1") Then 
				' User may log on the whole day
				DayLogon(DayCount) = "immer"
			Else
				' User has at least one exclusion, so walk through the hours
				' start counting at 1 so we can use the InStr 0 value
				' (but we need to adjust this in the code)
				HourCount = 1
				Do Until HourCount = 25
					' find times the user may not log on
					HourCount = InStr(HourCount, DayOfWeek(DayCount), "0")
					If HourCount = 0 Then 
						' no more exclusions, day complete
						HourCount = 25
					Else
						' add beginning of exclusion time
						DayLogon(DayCount) = DayLogon(DayCount) & ", nicht " & HourCount - 1
						' find next allowed hour
						HourCount = InStr(HourCount, DayOfWeek(DayCount), "1")
						If HourCount = 0 Then ' Or HourCount > 24
							' no more allowed hours, day complete
							HourCount = 25
							DayLogon(DayCount) = DayLogon(DayCount) & " bis 24"
						Else
							' add ending of exclusion time
							DayLogon(DayCount) = DayLogon(DayCount) & " bis " & HourCount - 1
							HourCount = HourCount + 1
						End If
					End If
				Loop
			End If	
			' Clear the beginning of the string
			If Left(DayLogon(DayCount), 2) = ", " Then DayLogon(DayCount) = Mid(DayLogon(DayCount), 3)
			DayCount = DayCount + 1
			Counter = 0
			LoopCounter = LoopCounter + 1
		End If
	Next
	If Join(DayOfWeek, "") = String(168, "1") Then
		' User may logon any time
		GetLogonHours = "jederzeit"
	ElseIf Join(DayOfWeek, "") = String(168, "0") Then
		' User may never log on
		GetLogonHours = "niemals"
	Else
		' Output Monday to Saturday
		For DayCount = 1 To 6
			GetLogonHours = GetLogonHours & DayNames(DayCount) & ": " & DayLogon(DayCount) & " <BR />" & vbNewLine
		Next
		' Outside the US, Sunday is the last day, not the first one
		GetLogonHours = GetLogonHours & DayNames(0) & ": " & DayLogon(0)
	End If
	
End Function

' ------------------------------------------------------------------------------
' Obtain local timezone bias from machine registry.
' ------------------------------------------------------------------------------
Function GetSystemBias
	Dim biasKey, k

	Dim objShell : Set objShell = CreateObject("WScript.Shell")
	biasKey = objShell.RegRead("HKLM\System\CurrentControlSet\Control\TimeZoneInformation\ActiveTimeBias")
	Set objShell = Nothing

	If (LCase(TypeName(biasKey)) = "long") Then
		GetSystemBias = biasKey
	ElseIf (LCase(TypeName(biasKey)) = "variant()") Then
		GetSystemBias = 0
		For k = 0 To UBound(biasKey)
			GetSystemBias = GetSystemBias + (biasKey(k) * 256^k)
		Next
	End If
End Function

' ------------------------------------------------------------------------------
' Convert Integer8 (64-bit) value to a date, adjusted for local time zone bias.
' ------------------------------------------------------------------------------
Function SetPermissionReason(ByVal Reason)
	If Not cfg("showAdvancedPermissionHint") Then Exit Function
	Dim i
	i = UBound(PermissionReason) + 1
	ReDim Preserve PermissionReason(i)
	PermissionReason(i) = Reason
End Function


' ------------------------------------------------------------------------------
' Escape illegal characters in a DN.
' It seems that the forward slash is the only character that is not auto-escaped.
' AD seems to accept a slash in an object name, however, using such a DN results
' in a bind error. That's why we need this function. Maybe it even has to be extended.
' ------------------------------------------------------------------------------
Function EscapeDN(ByVal DN)
	EscapeDN = Replace(DN, "/", "\/")
End Function

' ------------------------------------------------------------------------------
' Convert Integer8 (64-bit) value to a date, adjusted for local time zone bias.
' ------------------------------------------------------------------------------
Function Integer8Date(ByVal objDate, ByVal bias)
	Dim hi, lo
	hi = objDate.HighPart
	lo = objDate.LowPart
	' Account for bug in IADslargeInteger property methods.
	If (lo < 0) Then hi = hi + 1
	If (hi = 0) And (lo = 0) Then bias = 0
	On Error Resume Next
	Err.Clear
	Integer8Date = CDate(#1/1/1601# + (((hi * 2^32) + lo) / 600000000 - bias) / 1440)
	If Err.Number <> 0 Then Integer8Date = #1/1/1601#
	On Error Goto 0
End Function

' -------------------------------------------
' Convert Integer8 (64-bit) value to a number
' -------------------------------------------
Function Integer8Number(ByVal objInteger)
	Dim hi, lo
	hi = objInteger.HighPart
	lo = objInteger.LowPart
	Set objInteger = Nothing
	' Account for bug in IADslargeInteger property methods.
	If (lo < 0) Then hi = hi + 1
	If lo <> 0 Or hi <> -(2^31) Then
		Integer8Number = -(hi * 2^32 + lo) / 600000000
	End If
End Function

' ------------------------------------------------------------------------------
' Determine whether the given computer object is a domain controller or not.
' ------------------------------------------------------------------------------
Function GetDCRoles(objComputer)
	Dim options, objRID, objRS, objNtds

	' Only DCs have an RID set.
	' Attention: We might have to change this as it might not find RODCs!
	Set objRID = Nothing
	On Error Resume Next
	Set objRID = GetObject("LDAP://CN=RID Set," & objComputer.Get("distinguishedName"))
	On Error Goto 0

	If Not objRID Is Nothing Then
		' Computer is a domain controller.
		' Check if it's also a global catalog server.
		options = 0
		Set objRS = QueryAD("SELECT distinguishedName FROM 'LDAP://" _
			& rootDSE.Get("configurationNamingContext") & "' " _
			& "WHERE objectClass = 'server' and name='" & objComputer.Get("name") & "'")
		If Not objRS.EOF Then
			Set objNtds = GetObject("LDAP://CN=NTDS Settings," _
				& objRS("distinguishedName"))
			On Error Resume Next
			options = objNtds.Get("options")
			On Error Goto 0
		End If

		If options And NTDSDSA_OPT_IS_GC Then
			'
			GetDCRoles = Array("Domain Controller", "Global Catalog Server")
		Else
			GetDCRoles = "Domain Controller"
		End If
	Else
		' Computer is a domain member.
		GetDCRoles = "Domain member"
	End If
End Function

' ------------------------------------------------------------------------------
' Decode the TrustAttributes flags.
' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
' ------------------------------------------------------------------------------
Function DecodeTrustFlags(trustAttr)
	Dim flag, arrFlags, i

	DecodeTrustFlags = ""
	If Not IsNumeric(trustAttr) Then Exit Function

	arrFlags = Array()
	i = 0
	For Each flag In trustFlags.Keys
		If TrustAttr And flag Then
			ReDim Preserve arrFlags(i)
			arrFlags(i) = trustFlags(flag)
			i = i + 1
		End If
	Next

	DecodeTrustFlags = Join(arrFlags, ", ")
End Function

' ------------------------------------------------------------------------------
' Decode the UserAccountControl flags.
' See <http://msdn.microsoft.com/en-us/library/ms680832.aspx>
' ------------------------------------------------------------------------------
Function DecodeUACFlags(uac)
	Dim flag, arrFlags, i

	DecodeUACFlags = ""
	If Not IsNumeric(uac) Then Exit Function

	arrFlags = Array()
	i = 0
	For Each flag In uacFlags.Keys
		If uac And flag Then
			ReDim Preserve arrFlags(i)
			arrFlags(i) = uacFlags(flag)
			i = i + 1
		End If
	Next

	DecodeUACFlags = Join(arrFlags, ", ")
End Function

' ------------------------------------------------------------------------------
' Decode the pwdProperties flags.
' See <http://msdn.microsoft.com/en-us/library/cc245476(PROT.13).aspx>
' ------------------------------------------------------------------------------
Function DecodePwdProperties(pwdProp)

	DecodePwdProperties = ""
	If Not IsNumeric(pwdProp) Then Exit Function

	If pwdProp And DOMAIN_PASSWORD_COMPLEX Then DecodePwdProperties = "complex Passwords"
	If pwdProp And DOMAIN_PASSWORD_NO_CLEAR_CHANGE Then DecodePwdProperties = DecodePwdProperties & ", no clear-text Password changes"
	If pwdProp And DOMAIN_PASSWORD_STORE_CLEARTEXT Then DecodePwdProperties = DecodePwdProperties & ", Store passwords decrypted"
	
	If Left(DecodePwdProperties,2) = ", " Then DecodePwdProperties = Mid(DecodePwdProperties, 2)
	
	if trim(DecodePwdProperties) = "" then DecodePwdProperties = "no specifications"
	
End Function

' ------------------------------------------------------------------------------
' Translate a given canonical name into a "natural" one.
' ------------------------------------------------------------------------------
Function Translate(name)
	Translate = name
	If naturalName.Exists(LCase(name)) Then Translate = naturalName(LCase(name))
End Function

' ------------------------------------------------------------------------------
' Read GPO comment from file.
' ------------------------------------------------------------------------------
Function GetGPOComment(path)
	GetGPOComment = ""
	If path <> "" Then
		path = fso.BuildPath(path, "GPO.cmt")
		If fso.FileExists(path) Then
			Dim f: Set f = fso.OpenTextFile(path, ForReading, False, TristateUseDefault)
			GetGPOComment = f.ReadAll
			f.Close
			Set f = Nothing
		End If
		TrapError "GetGPOComment()"
	End If
End Function

' ------------------------------------------------------------------------------
' Get WMI filter (if present).
' ------------------------------------------------------------------------------
Function GetWMIFilter(adsPath)
	PrintDebug vbTab & "Checking WMI Filter for " & adsPath
	Dim objGPOWMI : Set objGPOWMI = GetObject(adsPath)
	GetWMIFilter = ""
	GetWMIFilter = objGPOWMI.Get("gpcWQLFilter")
	If GetWMIFilter <> "" Then GetWMIFilter = WMIFilterName(GetWMIFilter)
End Function

' ------------------------------------------------------------------------------
' Get the WMI filter name matching the given GUID name
' ------------------------------------------------------------------------------
Function WMIFilterName(WMIGUID)
	PrintDebug vbTab & "Getting WMI Filter friendly name for " & WMIGUID
	Dim WMIGUIDPos, WMIGUIDEnd
	Dim WMICont, objWMI
	' Isolate GUID
	WMIGUIDPos = InStr(WMIGUID, "{")
	WMIGUIDEnd = InStr(WMIGUID, "}")
	WMIGUID = Mid(WMIGUID, WMIGUIDPos + 1, WMIGUIDEnd - WMIGUIDPos - 1)
    ' Browse through WMI objects
	Set WMICont = GetObject("LDAP://CN=SOM,CN=WMIPolicy,CN=System," & rootDSE.Get("defaultNamingContext"))
    WMICont.filter = Array("msWMI-Som")
    For Each objWMI In WMICont
		If InStr(objWMI.name, WMIGUID) > 0 Then
			' Found it, now get its name
			WMIFilterName = objWMI.get("msWMI-Name")
        End If
    Next
    Set WMICont = Nothing
	PrintDebug vbTab & "Found WMI Filter name " & WMIFilterName

End Function

' ------------------------------------------------------------------------------
' Get links of the named GPO in the given structure.
' ------------------------------------------------------------------------------
Function GetGPOLinks(name, structure)
	Dim fname, query, i, dn

	fname = "GetGPOLinks(" & name & "," & structure & ")"
	query = "SELECT distinguishedName FROM 'LDAP://" & structure _
		& "' WHERE gPLink='*" & name & "*' ORDER BY name"

	Dim arrGPOLinks : arrGPOLinks = Array()
	Dim objGPOLinks : Set objGPOLinks = QueryAD(query)
	TrapError fname & " - Get links from AD"

	i = 0
	Do Until objGPOLinks.EOF
		' Store DN of GPO link in array if the object has a readable DN.
		Err.Clear
		dn = objGPOLinks("distinguishedName").Value
		If Not TrapError(fname & " - Read DN") Then
			ReDim Preserve arrGPOLinks(i)
			arrGPOLinks(i) = dn
			i = i + 1
		End If
		objGPOLinks.MoveNext
	Loop

	Set objGPOLinks = Nothing
	GetGPOLinks = arrGPOLinks
End Function

' ------------------------------------------------------------------------------
' Run the given WQL query against AD using the globally established ADO
' connection.
' ------------------------------------------------------------------------------
Function QueryAD(wql)
	Set QueryAD = Nothing
	adoCmd.CommandText = wql
	Dim adoRS : Set adoRS = adoCmd.Execute
	If Not TrapError("QueryAD(" & wql & ")") Then Set QueryAD = adoRS
End Function

' ------------------------------------------------------------------------------
' Return True if either dn is a substring of filter or vice versa. Otherwise
' return False.
' ------------------------------------------------------------------------------
Function IsMatching(dn, Filter)
	If (InStr(1, dn, Filter, vbTextCompare) + InStr(1, Filter, dn, vbTextCompare)) > 0 Then
		IsMatching = True
	Else
		IsMatching = False
	End If
End Function

' ------------------------------------------------------------------------------
' Return True if the given category is "computer", otherwise return False.
' ------------------------------------------------------------------------------
Function IsComputer(category)
	IsComputer = (LCase(Left(category, 38)) = "cn=computer,cn=schema,cn=configuration")
End Function

' ------------------------------------------------------------------------------
' Check if cleartext ADSI error messages are available.
' ------------------------------------------------------------------------------
Function CheckForADSIErrorMessages()
	CheckForADSIErrorMessages = False
	Dim ADError : Set ADError = Nothing
	On Error Resume Next
	Set ADError = CreateObject("ADsError")
	On Error Goto 0
	If Not ADError Is Nothing Then CheckForADSIErrorMessages = True
End Function

' ------------------------------------------------------------------------------
' Format a domain's distinguished name as an FQDN (dot-notation).
' ------------------------------------------------------------------------------
Function DN2FQDN(dn)
	DN2FQDN = Replace(Replace(dn, ",DC=", "."), "DC=", "")
End Function

' ------------------------------------------------------------------------------
' Get the name of the given FSMO role owner.
' ------------------------------------------------------------------------------
Function GetName(fSMORoleOwner)
	Dim objNTDS : Set objNtds = GetObject("LDAP://" & fSMORoleOwner)
	Dim objComputer : Set objComputer = GetObject(objNtds.Parent)
	GetName = Replace(objComputer.Name, "CN=", "")
	Set objNtds = Nothing
	Set objComputer = Nothing
End Function

' ------------------------------------------------------------------------------
' Convert the given octet array into a string representation of its integer
' equivalent.
' ------------------------------------------------------------------------------
Function Octet2Str(arr)
	Dim i

	Octet2Str = 0
	For i = LBound(arr) To UBound(arr)
		Octet2Str = Octet2Str * 256 + arr(i)
	Next

	Octet2Str = CStr(Octet2Str)
End Function

' ------------------------------------------------------------------------------
' Convert an SID from (raw) binary string format into human-readable "S-..."
' notation.
'
' The binary encoding of an SID S-a-b-c-d-e-f-... is aNbbbbbbccccddddeeeeffff...
' where the respective octets are:
'
' a       (1 octet, revision number)
' N       (1 octet, number of DWORDs following the first 8 octets)
' bbbbbb  (6 octets, 48-bit number "b" in big-endian order)
' cccc    (DWORD (4 octets), 32-bit number "c" in little-endian order)
' dddd    (DWORD (4 octets), 32-bit number "d" in little-endian order)
' eeee    (DWORD (4 octets), 32-bit number "e" in little-endian order)
' ffff    (DWORD (4 octets), 32-bit number "f" in little-endian order)
' ...
'
' See: <http://blogs.msdn.com/oldnewthing/archive/2004/03/15/89753.aspx>
'
' For a list of well-known SIDs see <http://support.microsoft.com/kb/243330>.
' ------------------------------------------------------------------------------
Function Raw2Str(rawSID)
	Dim octets, i

	' Convert binary string to octet array.
	ReDim octets(LenB(rawSid))
	For i = 1 To LenB(rawSid)
		octets(i-1) = AscB(MidB(rawSid, i, 1))
	Next

	' Convert octet array to (human-readable) SID string.
	Raw2Str = "S-" & CStr(octets(0)) & "-" & Octet2Str(Array(octets(2), octets(3), octets(4), octets(5), octets(6), octets(7)))
	For i = 8 To (4 * octets(1) + 4) Step 4
		Raw2Str = Raw2Str & "-" & Octet2Str(Array(octets(i+3), octets(i+2), octets(i+1), octets(i)))
	Next
End Function

' ------------------------------------------------------------------------------
' If an error has occurred: try to decode the error number, report both error
' number and error message, and return True. Otherwise return False.
' ------------------------------------------------------------------------------
Function TrapError(msg)
	TrapError = False
	If Err.number <> 0 Then
		If hasADSIErrorMessages Then
			Dim objError : Set objError = CreateObject("ADsError")
			msg = msg & ": [ADSI] " & objError.GetErrorMessage(Err.Number) _
				& " (0x" & Hex(Err.Number) & ")"
			Set objError = Nothing
		Else
			msg = msg & ": " & Err.Description & " (" & Err.Number & ")"
		End If
		WScript.StdErr.WriteLine msg
		PrintDebug msg
		Err.Clear
		TrapError = True
	End If
End Function

' ------------------------------------------------------------------------------
' Print the syntax of how the script has to be invoked.
' ------------------------------------------------------------------------------
Sub PrintSyntax()
	WScript.StdOut.WriteLine vbNewLine & cfg("generator")
	WScript.StdOut.WriteLine vbNewLine & "Syntax:" & vbNewLine _
		& "cscript.exe " & WScript.ScriptName & " [/d:<Definition>] [/r:<Report>] [/t:<Titel>]" & vbNewLine _
		& "  [/f:<Filter>] [/debug]" & vbNewLine & vbNewLine _
		& "  /d:    (optional) A Jos definition file name this" & vbNewLine _
		& "         parameter is not specified, the script reads the definitions" & vbNewLine _
		& "         from the standard input." & vbNewLine _
		& "  /r:    (optional) Name of the report file to be generated (HTML)" & vbNewLine _
		& "         Note: the file will be overwritten without question!" & vbNewLine _
		& "  /t:    (optional) Title of the Report" & vbNewLine _
		& "  /f:    (optional) LDAP object filter (Container/OU)" & vbNewLine _
		& "  /debug (optional) Switches on the debug mode." & vbNewLine & vbNewLine _
		& "For file name, specify only the file name (without path)!"
	WScript.StdOut.WriteLine vbNewLine & "Example:" & vbNewLine _
		& "cscript.exe " & WScript.ScriptName & " /d:""DomainData.txt"" /r:""MyDomainData.htm"" /t:""My Domain"" /f:""OU=IT,OU=Germany,DC=domain,DC=tld"""
	WScript.Quit 0
End Sub

' ------------------------------------------------------------------------------
' Print the given message to STDOUT if debugging is enabled.
' ------------------------------------------------------------------------------
Sub PrintDebug(msg)
	If cfg("debugEnabled") Then WScript.StdOut.WriteLine "[" & Time() & "] " & msg
End Sub
